from typing import Optional

import astropy.units as u
import numpy as np
from astropy.units import Quantity
from harvesters.core import Harvester

from ..core import Detector


class Camera(Detector):
    """Adapter for GenICam/GenTL cameras.

    Attributes:
        _nodes: The GenICam node map of the camera.
            This map can be used to access camera properties,
            see the `GenICam/GenAPI documentation <https://www.emva.org/standards-technology/genicam/>`_
            and the `Standard Features Naming Convention
            <https://www.emva.org/wp-content/uploads/GenICam_SFNC_2_3.pdf>` for more details.

            The node map should not be used to set properties that are available as properties in the Camera object,
            such as `duration` (exposure time), `width`, `height`, `binning`, etc.

            Also, the node map should not be used to set properties while the camera is fetching a frame (i. e.,
            between `trigger()` and calling `result()` on the returned concurrent.futures.Future object).

    Note:
        This class is a thin wrapper around the Harvesters module,
        which is a generic adapter for GenICam/GenTL cameras.

    Example:
        >>> camera = Camera(serial_number='12345678')
        >>> camera.start()
        >>> frame = camera.capture_frame()
        >>> camera.stop()
        """

    def __init__(self, cti_file: str, serial_number: Optional[str] = None, multi_threaded=True, **kwargs):
        """
            Initialize the Camera object.

            Args:
                cti_file: The path to the GenTL producer file.
                    This path depends on where the driver for the camera is installed.
                    For Basler cameras, this is typically located in
                    R"C:\\Program Files\\Basler\\pylon 7\\Runtime\\x64\\ProducerU3V.cti".

                serial_number: The serial number of the camera.
                    When omitted, the first camera found is selected.
                **kwargs: Additional keyword arguments.
                    These arguments are transferred to the node map of the camera.
        """
        self._harvester = Harvester()
        self._harvester.add_file(cti_file, check_validity=True)
        self._harvester.update()

        # open the camera, use the serial_number to select the camera if it is specified.
        search_key = {'serial_number': serial_number} if serial_number is not None else None
        self._camera = self._harvester.create(search_key=search_key)
        nodes = self._camera.remote_device.node_map

        print(dir(nodes))  # for debugging, should go in a separate function

        # set triggering to 'Software', so that we can trigger the camera by calling `trigger`.
        # turn off auto exposure so that `duration` accurately reflects the required measurement time.
        nodes.TriggerMode.value = 'On'
        nodes.TriggerSource.value = 'Software'
        nodes.ExposureMode.value = 'Timed'
        nodes.ExposureAuto.value = 'Off'
        nodes.BinningHorizontal.value = 1
        nodes.BinningVertical.value = 1
        nodes.OffsetX.value = 0
        nodes.OffsetY.value = 0
        nodes.Width.value = nodes.Width.max
        nodes.Height.value = nodes.Height.max
        self._nodes = nodes

        #  Todo:
        #         automatically expose a selection of properties in the node map as
        #         properties of the Camera object.
        #
        # assign a dynamic class, so that we can add properties to the Camera object
        # class DynamicClass(self.__class__):
        #     pass
        # self.__class__ = DynamicClass
        #
        # for key, value in kwargs.items():
        #     getter = lambda self: self.__dict__.get(key)
        #     setter = lambda self, val: self.__dict__.__setitem__(key, val)
        #     setattr(self.__class__, key, property(getter, setter))
        #     setattr(self, key, value)

        # set additional properties specified in the kwargs
        for key, value in kwargs.items():
            try:
                setattr(nodes, key, value)
            except AttributeError:
                print(f'Warning: could not set camera property {key} to {value}')

        try:
            pixel_size = [nodes.SensorPixelHeight.value, nodes.SensorPixelWidth.value] * u.um
        except AttributeError:  # the SensorPixelWidth feature is optional
            pixel_size = None

        super().__init__(multi_threaded=multi_threaded, data_shape=None, pixel_size=pixel_size, duration=None,
                         latency=0.0 * u.ms)
        self._camera.start()

    def __del__(self):
        if hasattr(self, '_camera'):
            self._camera.stop()
            self._camera.destroy()
        if hasattr(self, '_harvester'):
            self._harvester.reset()

    def _do_trigger(self):
        self._nodes.TriggerSoftware.execute()

    def paused(self):
        """Returns a context manager for pausing the camera.

        Usage ::
            >>> with camera.paused():
            >>>     camera.nodes.SomeNode.value = 10
        """
        return _CameraPause(self._camera)

    def _fetch(self, *args, **kwargs) -> np.ndarray:
        buffer = self._camera.fetch()
        frame = buffer.payload.components[0].data.reshape(self.data_shape)
        if frame.size == 0:
            raise Exception('Camera returned an empty frame')
        data = frame.copy()
        buffer.queue()  # give back buffer to the camera driver
        return data

    @property
    def duration(self) -> Quantity[u.ms]:
        """Returns the exposure time in milliseconds if software triggering is used.
        Returns ∞ if hardware triggering is used.
        TODO: implement hardware triggering."""
        return self.exposure_time.to(u.ms)

    @property
    def exposure_time(self) -> u.Quantity:
        """Exposure time in microseconds."""
        return self._nodes.ExposureTime.value * u.us

    @exposure_time.setter
    def exposure_time(self, value):
        with self.paused():
            self._nodes.ExposureTime.value = int(value.to(u.us).value)

    @property
    def binning(self) -> int:
        """Pixel binning factor

        Note:
            setting horizontal and vertical binning separately is not supported.
        """
        return self._nodes.BinningHorizontal.value

    @binning.setter
    def binning(self, value):
        with self.paused():
            if value != self._nodes.BinningHorizontal.value:
                self._nodes.BinningHorizontal.value = int(value)
            if value != self._nodes.BinningVertical.value:
                self._nodes.BinningVertical.value = int(value)

    @property
    def top(self) -> int:
        """
        Top position of the camera frame.

        Returns:
        int: The top position of the camera frame.
        """
        return self._nodes.OffsetY.value

    @top.setter
    def top(self, value: int):
        self._set_round_up(self._nodes.OffsetX, value)

    @property
    def left(self) -> int:
        """
        Left position of the camera frame.

        Returns:
        int: The left position of the camera frame.
        """
        return self._nodes.OffsetX.value

    @left.setter
    def left(self, value: int):
        self._set_round_up(self._nodes.OffsetX, value)

    def _set_round_up(self, node, value):
        """Sets the value of a property, rounding up to the next multiple of the increment."""
        inc = node.inc
        with self.paused():
            # round up value to the next multiple of inc
            node.value = int(value) + ((-value) % inc)

    @property
    def width(self) -> int:
        """
        Width of the camera frame.

        Returns:
        int: The width of the camera frame.
        """
        return self._nodes.Width.value

    @width.setter
    def width(self, value: int):
        self._set_round_up(self._nodes.Width, value)

    @property
    def height(self) -> int:
        """
        Height of the camera frame.

        Returns:
        int: The height of the camera frame.
        """
        return self._nodes.Height.value

    @height.setter
    def height(self, value: int):
        self._set_round_up(self._nodes.Height, value)

    @property
    def pixel_size(self) -> Optional[Quantity[u.um]]:
        """
        Physical pixel size of the camera sensor.

        Returns:
        Quantity: The pixel size of the camera.
        """
        return self._pixel_size

    @property
    def data_shape(self):
        return self.height, self.width

    @staticmethod
    def enumerate_cameras(cti_file: str):
        with Harvester() as harvester:
            harvester.add_file(cti_file, check_validity=True)
            harvester.update()
            return harvester.device_info_list.copy()


class _CameraPause:
    def __init__(self, camera):
        self._camera = camera

    def __enter__(self):
        return self._camera.stop()

    def __exit__(self, _type, _value, _traceback):
        self._camera.start()



from enum import Enum
from typing import Optional

import numpy as np
from numpy.typing import ArrayLike


class WFSResult:
    """
    Data structure for holding wavefront shaping results and statistics.

    Attributes:
        t (ndarray): Measured transmission matrix. If multiple targets were used, the first dimension(s) of `t`
            denote the columns of the transmission matrix (`a` indices), and the last dimensions(s) denote the targets,
            i. e., the rows of the transmission matrix (`b` indices).
        axis (int): Number of dimensions used for denoting a single column of the transmission matrix
            (e. g., 2 dimensions representing the x and y coordinates of the SLM pixels).
        fidelity_noise (ndarray): The estimated loss in fidelity caused by the limited SNR (for each target).
        fidelity_amplitude (ndarray): Estimated reduction of the fidelity due to phase-only modulation (for each target)
            (≈ π/4 for fully developed speckle).
        fidelity_calibration (ndarray): Estimated deviation from a sinusoid response.
        n (int): Total number of segments used in the optimization. When missing, this value is set to the number of
            elements in the first `axis` dimensions of `t`.
        estimated_optimized_intensity (ndarray): When missing, estimated intensity in the target(s) after displaying the
            wavefront correction on a perfect phase-only SLM.
        intensity_offset (Optional[ndarray]): Offset in the signal strength, as a scalar, or as one value per target.
            This is the offset that is caused by a bias in the detector signal, stray light, etc. Default value: 0.0.
    """

    def __init__(self,
                 t: np.ndarray,
                 t_f: np.ndarray,
                 axis: int,
                 fidelity_noise: ArrayLike,
                 fidelity_amplitude: ArrayLike,
                 fidelity_calibration: ArrayLike,
                 n: Optional[int] = None,
                 intensity_offset: Optional[ArrayLike] = 0.0):
        """
        Args:
            t(ndarray): measured transmission matrix.
            axis(int):
                number of dimensions used for denoting a single columns of the transmission matrix
                (e. g. 2 dimensions representing the x and y coordinates of the SLM pixels)
            fidelity_noise(ArrayLike):
                the estimated loss in fidelity caused by the the limited snr (for each target).
            fidelity_amplitude(ArrayLike):
                estimated reduction of the fidelity due to phase-only modulation (for each target)
                (≈ π/4 for fully developed speckle)
            fidelity_calibration(ArrayLike):
                estimated deviation from a sinusoid responds (TODO: experimental, untested)
            n(Optional[int]): total number of segments used in the optimization.
                when missing, this value is set to the number of elements in the first `axis` dimensions of `t`.
            intensity_offset(Optional[ArrayLike]):
                offset in the signal strength, as a scalar, or as one value per target
                this is the offset that is caused by a bias in the detector signal, stray light, etc.
                default value: 0.0

        """
        self.t = t
        self.t_f = t_f
        self.axis = axis
        self.fidelity_noise = np.atleast_1d(fidelity_noise)
        self.n = np.prod(t.shape[0:axis]) if n is None else n
        self.fidelity_amplitude = np.atleast_1d(fidelity_amplitude)
        self.fidelity_calibration = np.atleast_1d(fidelity_calibration)
        self.estimated_enhancement = np.atleast_1d(
            1.0 + (self.n - 1) * self.fidelity_amplitude * self.fidelity_noise * self.fidelity_calibration)
        self.intensity_offset = intensity_offset * np.ones(self.fidelity_calibration.shape) if np.isscalar(
            intensity_offset) \
            else intensity_offset
        after = np.sum(np.abs(t), tuple(
            range(self.axis))) ** 2 * self.fidelity_noise * self.fidelity_calibration + intensity_offset
        self.estimated_optimized_intensity = np.atleast_1d(after)

    def __str__(self) -> str:
        noise_warning = "OK" if self.fidelity_noise > 0.5 else "WARNING low signal quality."
        amplitude_warning = "OK" if self.fidelity_amplitude > 0.5 else "WARNING uneven contribution of optical modes."
        calibration_fidelity_warning = "OK" if self.fidelity_calibration > 0.5 else (
            "WARNING non-linear phase response, check "
            "lookup table.")
        return f"""
        Wavefront shaping results:
            fidelity_noise: {self.fidelity_noise} {noise_warning}
            fidelity_amplitude: {self.fidelity_amplitude} {amplitude_warning}
            fidelity_calibration: {self.fidelity_calibration} {calibration_fidelity_warning}
            estimated_enhancement: {self.estimated_enhancement}
            estimated_optimized_intensity: {self.estimated_optimized_intensity}
            """

    def select_target(self, b) -> 'WFSResult':
        """
        Returns the wavefront shaping results for a single target

        Args:
            b(int): target to select, as integer index.
                If the target array is multidimensional, it is flattened before selecting the `b`-th component.

        Returns: WFSResults data for the specified target
        """
        return WFSResult(t=self.t.reshape((*self.t.shape[0:2], -1))[:, :, b],
                         t_f=self.t_f.reshape((*self.t_f.shape[0:2], -1))[:, :, b],
                         axis=self.axis,
                         intensity_offset=self.intensity_offset[:][b],
                         fidelity_noise=self.fidelity_noise[:][b],
                         fidelity_amplitude=self.fidelity_amplitude[:][b],
                         fidelity_calibration=self.fidelity_calibration[:][b],
                         n=self.n,
                         )


def analyze_phase_stepping(measurements: np.ndarray, axis: int, A: Optional[float] = None):
    """Analyzes the result of phase stepping measurements, returning matrix `t` and noise statistics

    This function assumes that all measurements were made using the same reference field `A`
    and that the phase of the modulated segment/mode is phase-stepped in equally spaced steps
    between 0 and 2π, whereas the phase of the reference field is kept constant.

    Args:
        measurements(ndarray): array of phase stepping measurements.
            The array holds measured intensities
            with the first one or more dimensions corresponding to the segments(pixels) of the SLM,
            one dimension corresponding to the phase steps,
            and the last zero or more dimensions corresponding to the individual targets
            where the feedback was measured.
        axis(int): indicates which axis holds the phase steps.
        A(Optional[float]): magnitude of the reference field.
            This value is used to correctly normalize the returned transmission matrix.
            When missing, the value of `A` is estimated from the measurements.

    With `phase_steps` phase steps, the measurements are given by

    .. math::

        I_p = \\lvert A + B \\exp(i 2\\pi p / phase_{steps})\\rvert^2,

    This function computes the Fourier transform.

    .. math::

        \\frac{1}{phase_{steps}} \\sum I_p  \\exp(-i 2\\pi p / phase_{steps}) = A^* B

    The value of A^* B for each set of measurements is stored in the `field` attribute of the return
    value.
    Other attributes hold an estimate of the signal-to-noise ratio,
    and an estimate of the maximum enhancement that can be expected
    if these measurements are used for wavefront shaping.
    """
    phase_steps = measurements.shape[axis]
    n = int(np.prod(measurements.shape[:axis]))
    # M = np.prod(measurements.shape[axis + 1:])
    segments = tuple(range(axis))

    # Fourier transform the phase stepping measurements
    t_f_raw = np.fft.fft(measurements, axis=axis) / phase_steps

    if A is None:  # reference field strength not known: estimate from data
        t_abs = np.abs(np.take(t_f_raw, 1, axis=axis))
        offset = np.take(t_f_raw, 0, axis=axis)
        a_plus_b = np.sqrt(offset + 2.0 * t_abs)
        a_minus_b = np.sqrt(offset - 2.0 * t_abs)
        A = 0.5 * np.mean(a_plus_b + a_minus_b)

    t_f = t_f_raw / A
    t = np.take(t_f, 1, axis=axis)

    # compute the effect of amplitude variations.
    # for perfectly developed speckle, and homogeneous illumination, this factor will be pi/4
    amplitude_factor = np.mean(np.abs(t), segments) ** 2 / np.mean(np.abs(t) ** 2, segments)

    # estimate the calibration error
    # we first construct a matrix that can be used to fit
    # parameters a and b such that a·t(:) + b·t^*(:) ≈ t_f(:, k, :)
    ff = np.vstack((t.ravel(), np.conj(t.ravel()))).T
    ff_inv = np.linalg.pinv(ff)
    c = np.zeros(phase_steps, np.complex128)

    signal_energy = 0
    for k in range(1, phase_steps):
        cc = ff_inv @ np.take(t_f, k, axis=axis).ravel()
        signal_energy = signal_energy + np.sum(np.abs(ff @ cc) ** 2)
        c[k] = cc[0]

    # Estimate the error due to noise
    # The signal consists of the response with incorrect modulation,
    # (which occurs twice, ideally in the +1 and -1 components of the Fourier transform),
    # but this factor of two is already included in the 'signal_energy' calculation.
    # an offset, and the rest is noise.
    offset_energy = np.sum(np.take(t_f, 0, axis=axis) ** 2)
    total_energy = np.sum(np.abs(t_f) ** 2)

    if phase_steps > 3:
        noise_energy = (total_energy - signal_energy - offset_energy) / (phase_steps - 3)
        noise_factor = np.abs(np.maximum(signal_energy - noise_energy, 0.0) / signal_energy)
    else:
        noise_factor = 1.0  # cannot estimate reliably

    calibration_fidelity = np.abs(c[1]) ** 2 / np.sum(np.abs(c[1:]) ** 2)

    return WFSResult(t, t_f=t_f, axis=axis, fidelity_amplitude=amplitude_factor, fidelity_noise=noise_factor,
                     fidelity_calibration=calibration_fidelity, n=n)


class WFSController:
    """
    Controller for Wavefront Shaping (WFS) operations using a specified algorithm in the MicroManager environment.
    Manages the state of the wavefront and executes the algorithm to optimize and apply wavefront corrections, while
    exposing all these parameters to MicroManager.
    """

    class State(Enum):
        FLAT_WAVEFRONT = 0
        SHAPED_WAVEFRONT = 1

    def __init__(self, algorithm):
        """
        Args:
            algorithm: An instance of a wavefront shaping algorithm.
        """
        self.algorithm = algorithm
        self._wavefront = WFSController.State.FLAT_WAVEFRONT
        self._result = None
        self._noise_factor = None
        self._amplitude_factor = None
        self._estimated_enhancement = None
        self._calibration_fidelity = None
        self._estimated_optimized_intensity = None
        self._snr = None  # Average SNR. Computed when wavefront is computed.
        self._optimized_wavefront = None
        self._recompute_wavefront = False
        self._feedback_enhancement = None
        self._test_wavefront = False  # Trigger to test the optimized wavefront
        self._run_troubleshooter = False  # Trigger troubleshooter
        self.dark_frame = None
        self.before_frame = None

    @property
    def wavefront(self) -> State:
        """
        Gets the current wavefront state.

        Returns:
            State: The current state of the wavefront, either FLAT_WAVEFRONT or SHAPED_WAVEFRONT.
        """
        return self._wavefront

    @wavefront.setter
    def wavefront(self, value):
        """
        Sets the wavefront state and applies the corresponding phases to the SLM.

        Args:
            value (State): The desired state of the wavefront to set.
        """
        self._wavefront = value
        if value == WFSController.State.FLAT_WAVEFRONT:
            self.algorithm.slm.set_phases(0.0)
        else:
            if self._recompute_wavefront or self._optimized_wavefront is None:
                # select only the wavefront and statistics for the first target
                result = self.algorithm.execute().select_target(0)
                self._optimized_wavefront = -np.angle(result.t)
                self._noise_factor = result.fidelity_noise
                self._amplitude_factor = result.fidelity_amplitude
                self._estimated_enhancement = result.estimated_enhancement
                self._calibration_fidelity = result.fidelity_calibration
                self._estimated_optimized_intensity = result.estimated_optimized_intensity
                self._snr = 1.0 / (1.0 / result.fidelity_noise - 1.0)
                self._result = result
            self.algorithm.slm.set_phases(self._optimized_wavefront)

    @property
    def noise_factor(self) -> float:
        """
        Returns:
            float: noise factor: the estimated loss in fidelity caused by the limited snr.
        """
        return self._noise_factor

    @property
    def amplitude_factor(self) -> float:
        """
        Returns:
            float: amplitude factor: estimated reduction of the fidelity due to phase-only
            modulation (≈ π/4 for fully developed speckle)
        """
        return self._amplitude_factor

    @property
    def estimated_enhancement(self) -> float:
        """
        Returns:
            float: estimated enhancement: estimated ratio <after>/<before>  (with <> denoting
            ensemble average)
        """
        return self._estimated_enhancement

    @property
    def calibration_fidelity(self) -> float:
        """
        Returns:
            float: non-linearity.
        """
        return self._calibration_fidelity

    @property
    def estimated_optimized_intensity(self) -> float:
        """
        Returns:
            float: estimated optimized intensity.
        """
        return self._estimated_optimized_intensity

    @property
    def snr(self) -> float:
        """
        Gets the signal-to-noise ratio (SNR) of the optimized wavefront.

        Returns:
            float: The average SNR computed during wavefront optimization.
        """
        return self._snr

    @property
    def recompute_wavefront(self) -> bool:
        """Returns: bool that indicates whether the wavefront needs to be recomputed. """
        return self._recompute_wavefront

    @recompute_wavefront.setter
    def recompute_wavefront(self, value):
        """Sets the bool that indicates whether the wavefront needs to be recomputed. """
        self._recompute_wavefront = value

    @property
    def feedback_enhancement(self) -> float:
        """Returns: the average enhancement of the feedback, returns none if no such enhancement was measured."""
        return self._feedback_enhancement

    @property
    def test_wavefront(self) -> bool:
        """Returns: bool that indicates whether test_wavefront will be performed if set."""
        return self._test_wavefront

    @test_wavefront.setter
    def test_wavefront(self, value):
        """
        Calculates the feedback enhancement between the flat and shaped wavefronts by measuring feedback for both
        cases.

        Args:
            value (bool): True to enable test mode, False to disable.
        """
        if value:
            self.wavefront = WFSController.State.FLAT_WAVEFRONT
            feedback_flat = self.algorithm.feedback.read().copy()
            self.wavefront = WFSController.State.SHAPED_WAVEFRONT
            feedback_shaped = self.algorithm.feedback.read().copy()
            self._feedback_enhancement = float(feedback_shaped.sum() / feedback_flat.sum())

        self._test_wavefront = value



from typing import Optional

import astropy.units as u
import nidaqmx as ni
import nidaqmx.system
import numpy as np
from astropy.units import Quantity
from nidaqmx.constants import TaskMode
from nidaqmx.constants import TerminalConfiguration
from nidaqmx.stream_writers import AnalogMultiChannelWriter

from ..core import Detector
from ..utilities.patterns import coordinate_range


class ScanningMicroscope(Detector):
    """Controller for a laser-scanning microscope with two galvo mirrors
    controlled by a National Instruments data acquisition card (nidaq).

    Effectively, a `ScanningMicroscope` works like a camera, which can be triggered and returns 2-D images.
    These images are obtained by raster-scanning a focus using two galvo mirrors, controlled by a nidaq card,
    and recording a detector signal (typically from a photon multiplier tube (PMT)) with the same card.

    Upon construction, the maximum scan range (in volt) is set for both axes,
    and a scale parameter should be specified to convert these voltages to positions in the object plane.
    The number of pixels in the full scan range is specified with the `data_shape` parameter.

    After construction, a sub-region of the scan range (i. e., a region of interest ROI) may be selected using
    the `top`, `left`, `height` and `width` properties.
    Note that changing the ROI does not change the pixel size.
    Only the number of pixels in the returned image changes, as with a regular camera.

    The `binning` property can be used to change the pixel size without changing the ROI.
    By decreasing the `binning` property, more points are measured for the same ROI,
    thereby increasing the resolution.
    For compatibility with micromanager, the `binning` property should be an integer.
    A binning of 100 corresponds to the `data_shape` that is passed in the initializer of the object.

    The image quality at the edge of the scan range is usually low due to the fact
    that the mirror is rapidly changing the scan direction.
    The `padding` attribute corresponds to a fraction of the full voltage range along the fast axis (axis 1).
    This part of the data is discarded after the measurement.
    Note: the padding affects the physical size of the ROI, and hence the pixel_size
    The padding does not affect the number of pixels (`data_shape`) in the returned image.

    Example:
        >>> scanner = ScanningMicroscope(
                                bidirectional=True,
                                sample_rate=0.5 * u.MHz,
                                axis0=('Dev4/ao0', -2.0 * u.V, 2.0 * u.V),
                                axis1=('Dev4/ao1', -2.0 * u.V, 2.0 * u.V),
                                input=('Dev4/ai0', -1.0 * u.V, 1.0 * u.V),
                                data_shape=(1000, 1000),
                                scale=440 * u.um / u.V)
        >>> frame = scanner.read()
    """

    # LaserScanner(input=("ai/8", -1.0 * u.V, 1.0 * u.V))
    def __init__(self,
                 data_shape: tuple[int, ...],
                 input: tuple[str, Quantity[u.V], Quantity[u.V]],  # noqa
                 axis0: tuple[str, Quantity[u.V], Quantity[u.V]],
                 axis1: tuple[str, Quantity[u.V], Quantity[u.V]],
                 scale: Quantity[u.um / u.V],
                 sample_rate: Quantity[u.Hz],
                 delay: float = 0.0,
                 padding: float = 0.05,
                 binning: int = 1,
                 bidirectional: bool = True,
                 simulation: Optional[str] = None,
                 multi_threaded: bool = True,
                 preprocess: Optional[callable] = None):
        """
        Args:
            data_shape (tuple[int, int]): number of data points (height, width) in the full field of view.
                Note that the ROI can be reduced later by setting width, height, top and left,
                and the resolution can be changed by modifying the `binning` property.
            input: tuple[str, Quantity[u.V], Quantity[u.V]],
                 Description of the NI-DAQ channel to use for the input.
                 Tuple of: (name of the channel (e. g., 'ai/1'), minimum voltage, maximum voltage).
            axis0: tuple[str, Quantity[u.V], Quantity[u.V], TerminalConfiguration],
                 Description of the NI-DAQ channel to use for controlling the axis 0 galvo (slow axis).
                 The TerminalConfiguration element is optional and defaults to TerminalConfiguration.DEFAULT.
            axis1: tuple[str, Quantity[u.V], Quantity[u.V], TerminalConfiguration],
                 Description of the NI-DAQ channel to use for controlling the axis 1 galvo (fast axis).
                 The TerminalConfiguration element is optional and defaults to TerminalConfiguration.DEFAULT.
            scale (u.um / u.V):
                Conversion factor between voltage at the NI-DAQ card and displacement of the focus in the object plane.
                This may be an array of (height, width) conversion factors if the factors differ for the different axes.
                This factor is used to convert pixel positions `x_i, y_i` to voltages,
                 using the equation `(pos + [y_i, x_i]) * pixel_size / scale.
            sample_rate (u.Hz):
                Sample rate of the NI-DAQ input channel.
                The sample rate affects the total time needed to scan a single frame.
                Setting the ROI, padding, or binning does not affect the sample rate.
            delay (float): Delay between mirror control and data acquisition, measured in pixels
            padding (float): Padding fraction at the sides of the scan. The scanner will scan a larger area than will be
                reconstructed, to make sure the reconstructed image is within the linear scan range of the mirrors.
            bidirectional (bool): If true, enables bidirectional scanning along the fast axis.
            preprocess (callable): Process the raw data with this function before cropping. When None, the preprocessing
                will be skipped. The function must take input arguments data and sample_rate, and must return the
                preprocessed data.
        """
        # settings for input/output channels
        self._in_channel = input[0]
        self._in_v_min = input[1].to(u.V)
        self._in_v_max = input[2].to(u.V)
        self._in_terminal_configuration = input[3] if len(input) > 3 else TerminalConfiguration.DEFAULT
        self._axis0_channel = axis0[0]
        self._axis1_channel = axis1[0]
        self._out_v_min = Quantity((axis0[1], axis1[1])).to(u.V)
        self._out_v_max = Quantity((axis0[2], axis1[2])).to(u.V)
        self._sample_rate = sample_rate.to(u.Hz)

        self._scale = scale.repeat(2) if scale.size == 1 else scale

        v_width = (self._out_v_max - self._out_v_min)
        self._padding = np.array((0.0, float(padding)))
        self._v_origin = self._out_v_min + v_width * self._padding * 0.5  # Voltage corresponding to (top,left)=(0,0)
        self._roi_start = Quantity((0.0, 0.0), u.V)  # ROI start position relative to origin
        self._roi_end = v_width * (1.0 - self._padding)  # ROI end position relative to origin

        self._resized = True  # indicate that the output signals need to be recomputed
        self._binning = binning
        self._original_data_shape = data_shape

        # Scan settings
        self._delay = float(delay)
        self._padded_data_shape = np.array((0, 0))
        self._zoom = 1.0
        self._bidirectional = bidirectional

        self._write_task = None
        self._read_task = None

        self._valid = False  # indicates that `trigger()` should initialize the NI-DAQ tasks and scan pattern
        self._scan_pattern = None
        self._simulation = simulation

        self._original_pixel_size = ((self._out_v_max - self._out_v_min) * self._scale / data_shape).to(u.um)

        self._preprocess = preprocess

        # the pixel size and duration are computed dynamically
        # data_shape just returns self._data shape, and latency = 0.0 ms
        super().__init__(data_shape=data_shape, pixel_size=None, duration=None, latency=0.0 * u.ms,
                         multi_threaded=multi_threaded)
        self._update()

    def _update(self):
        # round padding up to integer number of pixels on both sides
        padding = 2 * np.ceil(self._data_shape * self._padding * 0.5).astype('int32')
        self._padded_data_shape = np.array(self._data_shape) + padding

        # compute the scan range.
        scan_range = (self._roi_end - self._roi_start) * (1.0 + padding / self._data_shape)  # noqa
        center = 0.5 * (self._roi_start + self._roi_end) + self._v_origin
        (voltages0, voltages1a) = coordinate_range(self._padded_data_shape, scan_range, offset=center)

        # clip to voltage limits to prevent possible damage
        voltages0 = voltages0.ravel().clip(self._out_v_min[0], self._out_v_max[0])
        voltages1a = voltages1a.ravel().clip(self._out_v_min[0], self._out_v_max[0])

        # for bidirectional scanning, reverse the direction of the even scan lines
        if self._bidirectional:
            voltages1b = voltages1a[::-1]
        else:
            voltages1b = voltages1a

        # Generate output voltages
        self._scan_pattern = np.zeros((2, *self._padded_data_shape))
        self._scan_pattern[0, :, :] = voltages0.to_value(u.V).reshape((-1, 1))
        self._scan_pattern[1, 0::2, :] = voltages1a.to_value(u.V).reshape((1, -1))
        self._scan_pattern[1, 1::2, :] = voltages1b.to_value(u.V).reshape((1, -1))
        self._scan_pattern = self._scan_pattern.reshape(2, -1)

        if self._simulation is not None:
            return
        # Sets up NI-DAQ task and i/o channels
        if self._read_task:
            self._read_task.close()
            self._read_task = None
        if self._write_task:
            self._write_task.close()
            self._write_task = None

        self._write_task = ni.Task()
        self._read_task = ni.Task()
        self._read_task.in_stream.timeout = self.timeout.to_value(u.s)

        # Configure the sample clock task
        sample_rate = self._sample_rate.to_value(u.Hz)
        sample_count = self._padded_data_shape[0] * self._padded_data_shape[1]

        # Configure the analog output task (two channels)
        self._write_task.ao_channels.add_ao_voltage_chan(self._axis0_channel,
                                                         min_val=self._out_v_min[0].to_value(u.V),
                                                         max_val=self._out_v_max[0].to_value(u.V))
        self._write_task.ao_channels.add_ao_voltage_chan(self._axis1_channel,
                                                         min_val=self._out_v_min[1].to_value(u.V),
                                                         max_val=self._out_v_max[1].to_value(u.V))
        self._write_task.timing.cfg_samp_clk_timing(sample_rate, samps_per_chan=sample_count)

        # Configure the analog input task (one channel)
        self._read_task.ai_channels.add_ai_voltage_chan(self._in_channel,
                                                        min_val=self._in_v_min.to_value(u.V),
                                                        max_val=self._in_v_max.to_value(u.V),
                                                        terminal_config=self._in_terminal_configuration)
        self._read_task.timing.cfg_samp_clk_timing(sample_rate, samps_per_chan=sample_count)
        self._read_task.triggers.start_trigger.cfg_dig_edge_start_trig(self._write_task.triggers.start_trigger.term)
        delay = (self._delay / self._sample_rate).to_value(u.s)
        if delay > 0.0:
            self._read_task.triggers.start_trigger.delay = delay
            self._read_task.triggers.start_trigger.delay_units = nidaqmx.constants.DigitalWidthUnits.SECONDS

        self._writer = AnalogMultiChannelWriter(self._write_task.out_stream)
        self._valid = True

    def _do_trigger(self):
        if not self._valid:
            self._update()

        if self._simulation is not None:
            return

        self._read_task.wait_until_done()
        self._write_task.wait_until_done()

        # write the samples to output in the x-y channels
        print(self._scan_pattern.size)
        self._writer.write_many_sample(self._scan_pattern)

        # Start the tasks
        self._read_task.start()  # waits for trigger coming from the write task
        self._write_task.start()

    def _raw_to_cropped(self, raw: np.ndarray) -> np.ndarray:
        """Converts the raw scanner data back into a 2-dimensional image.

        Because the scanner can return both signed and unsigned integers, both cases are accounted for.
        This function crops the data if padding was added, and it
        flips the even rows back if scanned in bidirectional mode.
        """
        start = (self._padded_data_shape - self._data_shape) // 2
        end = self._padded_data_shape - start

        # Change the data type into uint16 if necessary
        if raw.dtype == np.int16:
            # add 32768 to go from -32768-32767 to 0-65535
            cropped = raw.reshape(self._padded_data_shape).view('uint16')[start[0]:end[0], start[1]:end[1]] + 0x8000
        elif raw.dtype == np.uint16:
            cropped = raw.reshape(self._padded_data_shape)[start[0]:end[0], start[1]:end[1]]
        else:
            raise ValueError('Only int16 and uint16 data types are supported at the moment.')

        if self._bidirectional:
            cropped[1::2, :] = cropped[1::2, ::-1]

        return cropped

    def _fetch(self) -> np.ndarray:  # noqa
        """Reads the acquired data from the input task."""
        if self._simulation is None:
            raw = self._read_task.in_stream.read()
            self._read_task.stop()
            self._write_task.stop()
        elif self._simulation == 'horizontal':
            raw = (self._scan_pattern[1, :] * 1000.0).round().astype('int16')  # in mV
        elif self._simulation == 'vertical':
            raw = (self._scan_pattern[0, :] * 1000.0).round().astype('int16')
        else:
            raise ValueError(
                f"Invalid simulation option {self._simulation}. Should be 'horizontal', 'vertical', or 'None'")

        # Preprocess raw data if a preprocess function is set
        if self._preprocess is None:
            preprocessed_raw = raw
        elif callable(self._preprocess):
            preprocessed_raw = self._preprocess(data=raw, sample_rate=self._sample_rate)
        else:
            raise TypeError(f"Invalid type for {self._preprocess}. Should be callable or None.")
        return self._raw_to_cropped(preprocessed_raw)

    @property
    def preprocess(self):
        """The function to preprocess raw data before cropping."""
        return self._preprocess

    @preprocess.setter
    def preprocess(self, value: Optional[callable]):
        if not callable(value) and not (value is None):
            raise TypeError(f"Invalid type for {self._preprocess}. Should be callable or None.")
        self._preprocess = value

    @property
    def pixel_size(self) -> Quantity:
        """The size of a pixel in the object plane."""
        return self._original_pixel_size * self._binning / self._zoom

    @property
    def duration(self) -> Quantity[u.ms]:
        """Total duration of scanning for one frame."""
        return self._padded_data_shape[0] * self._padded_data_shape[1] * self.dwell_time

    @property
    def left(self) -> int:
        """The leftmost pixel of the Region of Interest (ROI) in the scan range."""
        return int(np.round(self._roi_start[1] * self._scale[1] / self.pixel_size[1]))

    @left.setter
    def left(self, value: int):
        self._roi_start[1] = value * self.pixel_size[1] / self._scale[1]
        self._valid = False

    @property
    def top(self) -> int:
        """The topmost pixel of the ROI in the scan range."""
        return int(np.round(self._roi_start[0] * self._scale[0] / self.pixel_size[0]))

    @top.setter
    def top(self, value: int):
        self._roi_start[0] = value * self.pixel_size[0] / self._scale[0]
        self._valid = False

    @property
    def height(self) -> int:
        """The number of pixels in the vertical dimension of the ROI."""
        return self.data_shape[0]

    @height.setter
    def height(self, value):
        self._data_shape = (int(value), int(self.data_shape[1]))
        self._roi_end = self._roi_start + self.pixel_size * self._data_shape / self._scale
        self._valid = False

    @property
    def width(self) -> int:
        """The number of pixels in the horizontal dimension of the ROI."""
        return self.data_shape[1]

    @width.setter
    def width(self, value):
        self._data_shape = (self.data_shape[0], int(value))
        self._roi_end = self._roi_start + self.pixel_size * self._data_shape / self._scale
        self._valid = False

    @property
    def dwell_time(self) -> Quantity[u.us]:
        """The time spent on each pixel during scanning."""
        return (1.0 / self._sample_rate).to(u.us)

    @dwell_time.setter
    def dwell_time(self, value: Quantity[u.us]):
        old = self._sample_rate
        self._sample_rate = (1.0 / value).to(u.Hz)
        self._delay = float(self._delay / old * self._sample_rate)
        self._valid = False

    @property
    def delay(self) -> float:
        """Delay between the control signal to the mirrors and the start of data acquisition."""
        return self._delay  # add unit

    @delay.setter
    def delay(self, value: float):
        self._delay = value
        self._valid = False

    @property
    def padding(self) -> float:
        """Fraction of the scan range at the edges to discard to reduce edge artifacts."""
        return float(self._padding[1])

    @padding.setter
    def padding(self, value: float):
        self._padding[1] = float(value)
        self._v_origin = self._out_v_min + (self._out_v_max - self._out_v_min) * self._padding * 0.5
        self._valid = False

    @property
    def bidirectional(self) -> bool:
        """Whether scanning is bidirectional along the fast axis."""
        return self._bidirectional

    @bidirectional.setter
    def bidirectional(self, value: bool):
        self._bidirectional = value
        self._valid = False

    @property
    def zoom(self) -> float:
        """Zoom factor.
        The zoom factor determines the pixel size relative to the original pixel size.
        The original pixel size is given by `_scale * (_out_v_max - _out_v_min) / _data_shape`
        When the zoom factor is changed,
        the center of the region of interest and the number of pixels in the data remain constant.
        """
        return self._zoom

    @zoom.setter
    def zoom(self, value: float):
        roi_center = 0.5 * (self._roi_end + self._roi_start)
        roi_width_old = self._roi_end - self._roi_start
        roi_width_new = roi_width_old * self._zoom / value
        self._roi_start = roi_center - 0.5 * roi_width_new
        self._roi_end = roi_center + 0.5 * roi_width_new
        self._zoom = float(value)
        self._valid = False

    @property
    def binning(self) -> int:
        """Undersampling factor.

        Increasing the binning reduces the number of pixels in the image while keeping dwell time the same.
        As a result, the total duration of a scan decreases.

        Note:
            This behavior is different from that of a real camera.
            No actual binning is performed, the scanner just takes fewer steps in x and y

        Note: the ROI is kept the same as much as possible.
            However, due to rounding, it may vary slightly.

        """
        return self._binning

    @binning.setter
    def binning(self, value: int):
        ratio = self._binning / value
        self._binning = value
        self._data_shape = tuple(int(s) for s in np.round(np.array(self._data_shape) * ratio))
        self._roi_end = self._roi_start + self.pixel_size * self._data_shape / self._scale
        self._valid = False

    @staticmethod
    def list_devices():
        return [d.name for d in nidaqmx.system.System().devices]



import numpy as np

from .utilities import analyze_phase_stepping, WFSResult
from ..core import Detector, PhaseSLM
from ..utilities.patterns import tilt


class FourierBase:
    """Base class definition for the Fourier algorithms as described in [1].

      This algorithm optimises the wavefront in a Fourier-basis. The modes that are tested are provided into a 'k-space'
      of which each 'k-vector' represents a certain angled wavefront that will be tested. (more detailed explanation is
      found in _get_phase_pattern).

      As described in [1], these modes are measured by interfering a certain mode on one half of the SLM with a
      'reference beam'. This is done by not modulating the other half of the SLM. In order to find a full corrective
      wavefront therefore, the experiment has to be repeated twice for each side of the SLM. Finally, the two wavefronts
      are combined.

      [1]: Bahareh Mastiani, Gerwin Osnabrugge, and Ivo M. Vellekoop,
      "Wavefront shaping for forward scattering," Opt. Express 30, 37436-37445 (2022)
      """

    def __init__(self, feedback: Detector, slm: PhaseSLM, slm_shape: tuple[int, int], k_left: np.ndarray,
                 k_right: np.ndarray, phase_steps: int = 4):
        """

        Args:
            feedback (Detector): The feedback source, usually a detector that provides measurement data.
            slm (PhaseSLM): slm object.
                The slm may have the `extent` property set to indicate the extent of the back pupil of the microscope
                objective in slm coordinates. By default, a value of 2.0, 2.0 is used (indicating that the pupil
                corresponds to a circle of radius 1.0 on the SLM). However, to prevent artefacts at the edges of the
                SLM,it may be overfilled, such that the `phases` image is mapped to an extent of e. g. (2.2, 2.2), i. e.
                10% larger than the back pupil.
            slm_shape (tuple[int, int]): The shape of the SLM patterns and transmission matrices.
            k_left (numpy.ndarray): 2-row matrix containing the y, and x components of the spatial frequencies
                used as a basis for the left-hand side of the SLM.
                The frequencies are defined such that a frequency of (1,0) or (0,1) corresponds to
                a phase gradient of -π to π over the back pupil of the microscope objective, which results in
                a displacement in the focal plane of exactly a distance corresponding to the Abbe diffraction limit.
            k_right (numpy.ndarray): 2-row matrix containing the y and x components of the spatial frequencies
                for the right-hand side of the SLM.
                The number of frequencies need not be equal for k_left and k_right.
            phase_steps (int): The number of phase steps for each mode (default is 4).
        """
        self._execute_button = False
        self.slm = slm
        self.feedback = feedback
        self.phase_steps = phase_steps
        self.k_left = k_left
        self.k_right = k_right
        self.slm_shape = slm_shape

    def execute(self) -> WFSResult:
        """
        Executes the FourierDualRef algorithm, computing the SLM transmission matrix.

        Returns:
            WFSResult: An object containing the computed SLM transmission matrix and related data.
        """
        # left side experiment
        data_left = self._single_side_experiment(self.k_left, 0)

        # right side experiment
        data_right = self._single_side_experiment(self.k_right, 1)

        # Compute transmission matrix (=field at SLM), as well as noise statistics
        results = self.compute_t(data_left, data_right, self.k_left, self.k_right)
        results.left = data_left
        results.right = data_right
        results.k_left = self.k_left
        results.k_right = self.k_right
        return results

    def _single_side_experiment(self, k_set: np.ndarray, side: int) -> WFSResult:
        """
        Conducts experiments on one side of the SLM, generating measurements for each spatial frequency and phase step.

        Args:
            k_set (np.ndarray): An array of spatial frequencies to use in the experiment.
            side (int): Indicates which side of the SLM to use (0 for the left hand side, 1 for right hand side).

        Returns:
            WFSResult: An object containing the computed SLM transmission matrix and related data.
        """
        measurements = np.zeros((k_set.shape[1], self.phase_steps, *self.feedback.data_shape))

        for i in range(k_set.shape[1]):
            for p in range(self.phase_steps):
                phase_offset = p * 2 * np.pi / self.phase_steps
                phase_pattern = self._get_phase_pattern(k_set[:, i], phase_offset, side)
                self.slm.set_phases(phase_pattern)
                self.feedback.trigger(out=measurements[i, p, ...])

        self.feedback.wait()
        return analyze_phase_stepping(measurements, axis=1)

    def _get_phase_pattern(self,
                           k: np.ndarray,
                           phase_offset: float,
                           side: int) -> np.ndarray:
        """
        Generates a phase pattern for the SLM based on the given spatial frequency, phase offset, and side.

        Args:
            k (np.ndarray): A 2-element array representing the spatial frequency.
            phase_offset (float): The phase offset to apply to the pattern.
            side (int): Indicates the side of the SLM for the pattern (0 for left, 1 for right).

        Returns:
            np.ndarray: The generated phase pattern.
        """
        # tilt generates a pattern from -2.0 to 2.0 (The convention for Zernike modes normalized to an RMS of 1).
        # The natural step to take is the Abbe diffraction limit, which corresponds to a gradient from
        # -π to π.
        num_columns = int(0.5 * self.slm_shape[1])
        tilted_front = tilt([self.slm_shape[0], num_columns], k * (0.5 * np.pi),
                            phase_offset=phase_offset)

        # Handle side-dependent pattern

        empty_part = np.zeros((self.slm_shape[0], self.slm_shape[1] - num_columns))

        # Concatenate based on the side
        if side == 0:
            # Place the pattern on the left
            result = np.concatenate((tilted_front, empty_part), axis=1)
        else:
            # Place the pattern on the right
            result = np.concatenate((empty_part, tilted_front), axis=1)

        return result

    def compute_t(self, left: WFSResult, right: WFSResult, k_left, k_right) -> WFSResult:
        """
        Computes the SLM transmission matrix by combining the Fourier transmission matrices from both sides of the SLM.

        Args:
            left (WFSResult): The wavefront shaping result for the left side.
            right (WFSResult): The wavefront shaping result for the right side.
            k_left (np.ndarray): The spatial frequency matrix for the left side.
            k_right (np.ndarray): The spatial frequency matrix for the right side.

        Returns:
            WFSResult: An object containing the combined transmission matrix and related statistics.
        """

        # TODO: determine noise
        # Initialize transmission matrices
        t1 = np.zeros((*self.slm_shape, *self.feedback.data_shape), dtype='complex128')
        t2 = np.zeros((*self.slm_shape, *self.feedback.data_shape), dtype='complex128')

        # Calculate phase difference between the two halves
        # We have two phase stepping measurements where both halves are flat (k=0)
        # Locate these measurements, and use the phase difference between them to connect the two halves
        # of the corrected wavefront.

        # Find the index of the (0,0) mode in k_left and k_right
        index_0_left = np.argmin(k_left[0] ** 2 + k_left[1] ** 2)
        index_0_right = np.argmin(k_right[0] ** 2 + k_left[1] ** 2)
        if not np.all(k_left[:, index_0_left] == 0.0) or not np.all(k_right[:, index_0_right] == 0.0):
            raise Exception("k=(0,0) component missing from the measurement set, cannot determine relative phase.")

        # average the measurements for better accuracy
        # TODO: absolute values are not the same in simulation, 'A' scaling is off?
        relative = 0.5 * (left.t[index_0_left, ...] + np.conjugate(right.t[index_0_right, ...]))

        # Apply phase correction to the right side
        phase_correction = relative / np.abs(relative)

        # Construct the transmission matrices
        normalisation = 1.0 / (0.5 * self.slm_shape[0] * self.slm_shape[1])
        for n, t in enumerate(left.t):
            phi = self._get_phase_pattern(k_left[:, n], 0, 0)
            t1 += np.tensordot(np.exp(-1j * phi), t * normalisation, 0)

        for n, t in enumerate(right.t):
            phi = self._get_phase_pattern(k_right[:, n], 0, 1)
            t2 += np.tensordot(np.exp(-1j * phi), t * (normalisation * phase_correction), 0)

        # Combine the left and right sides
        t_full = np.concatenate([t1[:, :self.slm_shape[0] // 2, ...], t2[:, self.slm_shape[0] // 2:, ...]], axis=1)
        t_f_full = np.concatenate([left.t_f, right.t_f],
                                  axis=1)  # also store raw data (not normalized or corrected yet!)

        # return combined result, along with a course estimate of the snr and expected enhancement
        # TODO: not accurate yet
        # for the estimated_improvement, first convert to field improvement, then back to intensity improvement
        def weighted_average(x_left, x_right):
            return (left.n * x_left + right.n * x_right) / (left.n + right.n)

        return WFSResult(t=t_full,
                         t_f=t_f_full,
                         n=left.n + right.n,
                         axis=2,
                         fidelity_noise=weighted_average(left.fidelity_noise, right.fidelity_noise),
                         fidelity_amplitude=weighted_average(left.fidelity_amplitude, right.fidelity_amplitude),
                         fidelity_calibration=weighted_average(left.fidelity_calibration, right.fidelity_calibration))



from .fourier import FourierBase
from ..core import Detector, PhaseSLM
import numpy as np


class FourierDualReference(FourierBase):
    """Fourier double reference algorithm, based on Mastiani et al. [1].

    It constructs a symmetric k-space for the algorithm.
    The k-space initializer is set to None because for custom k-spaces, you should use FourierDualRef directly.

    [1]: Bahareh Mastiani, Gerwin Osnabrugge, and Ivo M. Vellekoop,
    "Wavefront shaping for forward scattering," Opt. Express 30, 37436-37445 (2022)
    """

    def __init__(self, feedback: Detector, slm: PhaseSLM, slm_shape=(500, 500), phase_steps=4, k_angles_min=-3,
                 k_angles_max=3):
        """
        Args:
            feedback (Detector): Source of feedback
            slm (PhaseSLM): The spatial light modulator
            slm_shape (tuple of two ints): The shape that the SLM patterns & transmission matrices are calculated for,
                does not necessarily have to be the actual pixel dimensions as the SLM.
            phase_steps (int): The number of phase steps.
            k_angles_min (int): The minimum k-angle.
            k_angles_max (int): The maximum k-angle.
        """
        super().__init__(feedback, slm, slm_shape, np.array((0, 0)), np.array((0, 0)), phase_steps=phase_steps)
        self._k_angles_min = k_angles_min
        self._k_angles_max = k_angles_max

        self._build_k_space()

    def _build_k_space(self):
        """
        Constructs the k-space by creating Cartesian products of k_x and k_y angles.
        Fills the k_left and k_right matrices with the same k-space.

        Returns:
            None: The function updates the instance attributes.
        """
        kx_angles = np.arange(self._k_angles_min, self._k_angles_max + 1, 1)
        ky_angles = np.arange(self._k_angles_min, self._k_angles_max + 1, 1)
        # Make the cartesian product of kx_angles and ky_angles to make a square k-space

        k_x = np.repeat(np.array(kx_angles)[np.newaxis, :], len(ky_angles), axis=0).flatten()
        k_y = np.repeat(np.array(ky_angles)[:, np.newaxis], len(ky_angles), axis=1).flatten()
        self.k_left = np.vstack((k_x, k_y))
        self.k_right = np.vstack((k_x, k_y))

    @property
    def k_angles_min(self) -> int:
        """The lower bound of the range of angles in x and y direction"""
        return self._k_angles_min

    @k_angles_min.setter
    def k_angles_min(self, value):
        """Sets the lower bound of the range of angles in x and y direction, triggers the building of the internal
            k-space properties.
        """
        self._k_angles_min = value
        self._build_k_space()

    @property
    def k_angles_max(self) -> int:
        """The higher bound of the range of angles in x and y direction"""
        return self._k_angles_max

    @k_angles_max.setter
    def k_angles_max(self, value):
        """Sets the higher bound of the range of angles in x and y direction, triggers the building of the internal
            k-space properties."""
        self._k_angles_max = value
        self._build_k_space()



import time
from typing import Sequence, Optional

import astropy.units as u
import numpy as np
from astropy.units import Quantity

from ..core import Detector, Processor, Actuator
from ..processors import CropProcessor
from ..utilities import ExtentType, get_pixel_size


class StaticSource(Detector):
    """
    Detector that returns pre-set data. Also simulates latency and measurement duration.
    """

    def __init__(self, data: np.ndarray, pixel_size: Optional[ExtentType] = None, extent: Optional[ExtentType] = None,
                 latency: Quantity[u.ms] = 0 * u.ms, duration: Quantity[u.ms] = 0 * u.ms, multi_threaded: bool = None):
        """
        Initializes the MockSource
        TODO: factor out the latency and duration into a separate class?
        Args:
            data (np.ndarray): The pre-set data to be returned by the mock source.
            pixel_size (Quantity, optional): The size of each pixel in the data.
                If not specified, the pixel size is calculated from the extent and the data shape.
                If neither pixel size nor extent are specified, the pixel size from `data` is used, if available.
                Otherwise, pixel_size is set to None.
            extent (Optional[ExtentType]): The physical extent of the data array.
                Only used when the pixel size is not specified explicitly.
        """
        if pixel_size is None:
            if extent is not None:
                pixel_size = Quantity(extent) / data.shape
            else:
                pixel_size = get_pixel_size(data)

        if pixel_size is not None and (np.isscalar(pixel_size) or pixel_size.size == 1) and data.ndim > 1:
            pixel_size = pixel_size.repeat(data.ndim)

        if multi_threaded is None:
            multi_threaded = latency > 0 * u.ms or duration > 0 * u.ms

        self._data = data
        super().__init__(data_shape=data.shape, pixel_size=pixel_size, latency=latency, duration=duration,
                         multi_threaded=multi_threaded)

    def _fetch(self) -> np.ndarray:  # noqa
        total_time_s = self.latency.to_value(u.s) + self.duration.to_value(u.s)
        if total_time_s > 0.0:
            time.sleep(total_time_s)
        return self._data

    @property
    def data(self):
        """
        The pre-set data to be returned by the mock source.

        Note:
            The data is not copied.
            When setting the `data` property, the `data_shape` attribute is updated accordingly.
            If the `data` property is set with an array that has the `pixel_size` metadata set,
            the `pixel_size` attribute is also updated accordingly.
        """
        return self._data

    @data.setter
    def data(self, value):
        self._data = value
        self._pixel_size = get_pixel_size(value)
        self._data_shape = value.shape


class NoiseSource(Detector):
    def __init__(self, noise_type: str, *, data_shape: tuple[int, ...], pixel_size: Quantity, multi_threaded=True,
                 generator=None,
                 **kwargs):
        self._noise_type = noise_type
        self._noise_arguments = kwargs
        self._rng = generator if generator is not None else np.random.default_rng()
        super().__init__(data_shape=data_shape, pixel_size=pixel_size, latency=0 * u.ms, duration=0 * u.ms,
                         multi_threaded=multi_threaded)

    def _fetch(self) -> np.ndarray:  # noqa
        if self._noise_type == 'uniform':
            return self._rng.uniform(**self._noise_arguments, size=self.data_shape)
        elif self._noise_type == 'gaussian':
            return self._rng.normal(**self._noise_arguments, size=self.data_shape)
        else:
            raise ValueError(f'Unknown noise type: {self._noise_type}')

    @Detector.data_shape.setter
    def data_shape(self, value):
        self._data_shape = tuple(value)


class ADCProcessor(Processor):
    """Mimics an analog-digital converter.

    At the moment, only positive input and output values are supported.
    """

    def __init__(self, source: Detector, analog_max: float = 0.0, digital_max: int = 0xFFFF,
                 shot_noise: bool = False, gaussian_noise_std: float = 0.0, multi_threaded: bool = True,
                 generator=None):
        """
        Initializes the ADCProcessor class, which mimics an analog-digital converter.

        Args:
            source (Detector): The source detector providing analog data.
            analog_max (float): The maximum analog value that can be handled by the ADC.
                If set to 0.0, each measurement will be automatically scaled so that the maximum
                value in the data set returned by `source` is converted to `digital_max`.
                Note that this means that the values from two different measurements cannot be compared quantitatively.

            digital_max (int):
                The maximum digital value that the ADC can output, default is unsigned 16-bit maximum.

            shot_noise (bool):
                Flag to determine if Poisson noise should be applied instead of rounding.
                Useful for realistically simulating detectors.

            gaussian_noise_std (float):
                If >0, add gaussian noise with std of this value to the data.
        """
        super().__init__(source, multi_threaded=multi_threaded)
        self._analog_max = None
        self._digital_max = None
        self._shot_noise = None
        self._gaussian_noise_std = None
        self._rng = generator if generator is not None else np.random.default_rng()
        self.gaussian_noise_std = gaussian_noise_std
        self.shot_noise = shot_noise
        self.analog_max = analog_max  # check value
        self.digital_max = digital_max  # check value

    def _fetch(self, data) -> np.ndarray:  # noqa
        """Clips the data to the range of the ADC, and digitizes the values."""

        if self.analog_max == 0.0:  # auto scaling
            max_value = np.max(data)
            if max_value > 0.0:
                data = data * (self.digital_max / max_value)  # auto-scale to maximum value
        else:
            data = data * (self.digital_max / self.analog_max)

        if self.shot_noise:
            data = self._rng.poisson(data)

        if self._gaussian_noise_std > 0.0:
            data = data + self._rng.normal(scale=self._gaussian_noise_std, size=data.shape)

        return np.clip(np.rint(data), 0, self.digital_max).astype('uint16')

    @property
    def analog_max(self) -> Optional[float]:
        """Maximum value that the ADC can handle as input

        This value and all higher values are converted to `digital_max`.
        When set to 0.0, the input signal is scaled automatically so that the maximum corresponds to
        `digital_max`
        """
        return self._analog_max

    @analog_max.setter
    def analog_max(self, value):
        if value < 0.0:
            raise ValueError('analog_max cannot be negative')
        self._analog_max = value

    @property
    def digital_max(self) -> int:
        """Maximum value that the ADC can output.

        Default value is 0xFFFF (16 bits)
        """
        return self._digital_max

    @digital_max.setter
    def digital_max(self, value):
        if value < 0 or value > 0xFFFF:
            raise ValueError('digital_max must be between 0 and 0xFFFF')
        self._digital_max = int(value)

    @property
    def shot_noise(self) -> bool:
        """when True, apply Poisson noise to the data instead of rounding"""
        return self._shot_noise

    @shot_noise.setter
    def shot_noise(self, value: bool):
        self._shot_noise = value

    @property
    def gaussian_noise_std(self) -> int:
        return self._gaussian_noise_std

    @gaussian_noise_std.setter
    def gaussian_noise_std(self, value: int):
        self._gaussian_noise_std = value


class Camera(ADCProcessor):
    """Wraps any 2-D image source as a camera.

    To implement the camera interface, in addition to the Detector interface,
    we must implement left,right,top, and bottom.
    In addition, the data should be returned as uint16.
    Conversion to uint16 is implemented in the ADCProcessor base class.
    """

    def __init__(self, source: Detector, shape: Optional[Sequence[int]] = None,
                 pos: Optional[Sequence[int]] = None, **kwargs):
        """
        Args:
            source (Detector): The source detector to be wrapped.
            shape (Optional[Sequence[int]]): The shape of the image data to be captured.
            pos (Optional[Sequence[int]]): The position on the source from where the image is captured.
            **kwargs: Additional keyword arguments to be passed to the Detector base class.

            TODO: move left-right-top-bottom to CropProcessor.
            Expose the properties of CropProcessor as properties of MockCamera automatically by copying from __dict__?
        """
        self._crop = CropProcessor(source, shape=shape, pos=pos)
        super().__init__(source=self._crop, **kwargs)

    @property
    def left(self) -> int:
        """left (int): Horizontal start position of the ROI."""
        return self._crop.pos[1]

    @left.setter
    def left(self, value: int):
        self._crop.pos = (self._crop.pos[0], value)

    @property
    def right(self) -> int:
        """right (int): Horizontal end position of the ROI."""
        return self.left + self.width

    @property
    def top(self) -> int:
        """top (int): Vertical start position of the ROI."""
        return self._crop.pos[0]

    @top.setter
    def top(self, value: int):
        self._crop.pos = (value, self._crop.pos[1])

    @property
    def bottom(self) -> int:
        """bottom (int): Vertical end position of the ROI."""
        return self.top + self.height

    @property
    def height(self) -> int:
        """Height of the ROI."""
        return self._crop.data_shape[0]

    @height.setter
    def height(self, value: int):
        self._crop.data_shape = (value, self._crop.data_shape[1])

    @property
    def width(self) -> int:
        """Width of the ROI."""
        return self._crop.data_shape[1]

    @width.setter
    def width(self, value: int):
        self._crop.data_shape = (self._crop.data_shape[0], value)

    @property
    def data_shape(self):
        return self._crop.data_shape

    @data_shape.setter
    def data_shape(self, value):
        self._crop.data_shape = value


class XYStage(Actuator):
    """
    Mimics an XY stage actuator
    """

    def __init__(self, step_size_x: Quantity[u.um], step_size_y: Quantity[u.um]):
        """

        Args:
            step_size_x (Quantity[u.um]): The step size in the x-direction.
            step_size_y (Quantity[u.um]): The step size in the y-direction.
        """
        super().__init__(duration=0 * u.ms, latency=0 * u.ms)
        self.step_size_x = step_size_x.to(u.um)
        self.step_size_y = step_size_y.to(u.um)
        self._y = 0.0 * u.um
        self._x = 0.0 * u.um

    @property
    def x(self) -> Quantity[u.um]:
        return self._x

    @x.setter
    def x(self, value: Quantity[u.um]):
        self._x = value.to(u.um)

    @property
    def y(self) -> Quantity[u.um]:
        return self._y

    @y.setter
    def y(self, value: Quantity[u.um]):
        self._y = value.to(u.um)

    def home(self):
        self._x = 0.0 * u.um
        self._y = 0.0 * u.um


class Shutter(Processor):
    """
    A mock version of a shutter.
    When open, passes through the input field
    When closed, passes through the input field * 0.0
    """

    def __init__(self, source: Detector):
        super().__init__(source, multi_threaded=False)
        self._open = True

    @property
    def open(self) -> bool:
        return self._open

    @open.setter
    def open(self, value: bool):
        self._open = value

    def _fetch(self, source: np.ndarray) -> np.ndarray:  # noqa
        return source if self._open else 0.0 * source



import logging
import time

import matplotlib.pyplot as plt
import numpy as np
from numpy.fft import fft2, ifft2

from ..core import Detector, PhaseSLM


def signal_std(signal_with_noise: np.ndarray, noise: np.ndarray) -> float:
    """
    Compute noise corrected standard deviation of signal measurement.

    Args:
        signal_with_noise:
            ND array containing the measured signal including noise. The noise is assumed to be uncorrelated with the
            signal, such that var(measured) = var(signal) + var(noise).
        noise:
            ND array containing only noise.

    Returns:
        Standard deviation of the signal, corrected for the variance due to given noise. If the noise variance
        is greater than the signal_with_noise variance, the output will be 0.
    """
    signal_var = signal_with_noise.var() - noise.var()
    if signal_var < 0:
        signal_var = 0
    return np.sqrt(signal_var)


def cnr(signal_with_noise: np.ndarray, noise: np.ndarray) -> np.float64:
    """
    Compute the noise-corrected contrast-to-noise ratio of a measured signal. Contrast is computed as the standard
    deviation, corrected for noise. The noise variance is computed from a separate array, containing only noise.

    Args:
        signal_with_noise:
            ND array containing the measured signal including noise. The noise is assumed to be uncorrelated with the
            signal, such that var(measured) = var(signal) + var(noise).
        noise:
            ND array containing only noise, e. g. a dark frame.

    Returns:
        Standard deviation of the signal, corrected for the variance due to given noise.
    """
    return signal_std(signal_with_noise, noise) / noise.std()


def contrast_enhancement(signal_with_noise: np.ndarray, reference_with_noise: np.ndarray,
                         noise: np.ndarray) -> float:
    """
    Compute noise corrected contrast enhancement. The noise is assumed to be uncorrelated with the signal, such that
    var(measured) = var(signal) + var(noise).

    Args:
        signal_with_noise:
            ND array containing the measured signal including noise, e. g. image signal with shaped wavefront.
        reference_with_noise:
            ND array containing a reference signal including noise, e. g. image signal with a flat wavefront.
        noise:
            ND array containing only noise.

    Returns:
        Standard deviation of the signal, corrected for the variance due to given noise.
    """
    signal_std_sig = signal_std(signal_with_noise, noise)
    signal_std_ref = signal_std(reference_with_noise, noise)
    return signal_std_sig / signal_std_ref


def cross_corr_fft2(f: np.ndarray, g: np.ndarray) -> np.ndarray:
    """
    Compute cross-correlation with a 2D Fast Fourier Transform. Note that this approach will introduce wrap-around
    artefacts.

    Args:
        f, g:
            2D arrays to be correlated.
    """
    return ifft2(fft2(f).conj() * fft2(g))


def find_pixel_shift(f: np.ndarray, g: np.ndarray) -> tuple[int, ...]:
    """
    Find the pixel shift between two images by performing a 2D FFT based cross-correlation.
    """
    corr = cross_corr_fft2(f, g)  # Compute cross-correlation with fft2
    s = np.array(corr).shape  # Get shape
    index = np.unravel_index(np.argmax(corr), s)  # Find 2D indices of maximum
    # convert indices to pixel shift
    dy = index[0] if index[0] < s[0] / 2 else index[0] - s[0]
    dx = index[1] if index[1] < s[1] / 2 else index[1] - s[1]
    return dy, dx


def field_correlation(a: np.ndarray, b: np.ndarray) -> float:
    """
    Compute field correlation, i. e. inner product of two fields, normalized by the product of the L2 norms,
    such that field_correlation(f, s*f) == 1, where s is a scalar value.
    Also known as normalized first order correlation :math:`g_1`.

    Args:
        A, B    Real or complex fields (or other arrays) to be correlated.
                A and B must have the same shape.
    """
    return np.vdot(a, b) / np.sqrt(np.vdot(a, a) * np.vdot(b, b))


def frame_correlation(a: np.ndarray, b: np.ndarray) -> float:
    """
    Compute frame correlation between two frames.
    This is the normalized second order correlation :math:`g_2`.
    Note: This function computes the correlation between two frames. The output is a single value.

    Args:
        a, b    Real-valued intensity arrays of the (such as camera frames) to be correlated.
                A and B must have the same shape.

    This correlation function can be used to estimate the lowered fidelity due to speckle decorrelation.
    See also:
         [Jang et al. 2015] https://opg.optica.org/boe/fulltext.cfm?uri=boe-6-1-72&id=306198
    """
    return np.mean(a * b) / (np.mean(a) * np.mean(b)) - 1


def pearson_correlation(a: np.ndarray, b: np.ndarray, noise_var: np.ndarray = 0.0) -> float:
    """
    Compute Pearson correlation.

    The variances in the normalization factor can optionally be compensated for uncorrelated noise. This is done
    by subtracting the noise variance from the signal variance.

    Args:
        a, b: Real valued arrays.
        noise_var: Variance of uncorrelated noise to compensate for.
    """
    a_dev = a - a.mean()  # Deviations from mean a
    b_dev = b - b.mean()  # Deviations from mean b
    covar = (a_dev * b_dev).mean()  # Covariance
    a_var_signal = a.var() - noise_var  # Variance of signal in a, excluding noise
    b_var_signal = b.var() - noise_var  # Variance of signal in b, excluding noise
    return covar / np.sqrt(a_var_signal * b_var_signal)


class StabilityResult:
    """
    Result of a stability measurement.

    Attributes:
        pixel_shifts_first: Image shift in pixels, compared to first frame.
        correlations_first: Pearson correlations with first frame.
        correlations_disattenuated_first: Pearson correlations with first frame. Compensated for noise.
        contrast_ratios_first: Contrast ratio with first frame.
        pixel_shifts_prev: Image shift in pixels, compared to previous frame.
        correlations_prev: Pearson correlations with previous frame.
        correlations_disattenuated_prev: Pearson correlations with previous frame. Compensated for noise.
        contrast_ratios_prev: Contrast ratio with previous frame.
        timestamps: Timestamps in seconds since start of measurement.
        framestack: 3D array containing all recorded frames. Is None unless saving frames was requested.
    """

    def __init__(self, pixel_shifts_first, correlations_first, correlations_disattenuated_first, contrast_ratios_first,
                 pixel_shifts_prev, correlations_prev, correlations_disattenuated_prev, contrast_ratios_prev,
                 abs_timestamps, framestack):
        # Comparison with first frame
        self.pixel_shifts_first = pixel_shifts_first
        self.correlations_first = correlations_first
        self.correlations_disattenuated_first = correlations_disattenuated_first
        self.contrast_ratios_first = contrast_ratios_first

        # Comparison with previous frame
        self.pixel_shifts_prev = pixel_shifts_prev
        self.correlations_prev = correlations_prev
        self.correlations_disattenuated_prev = correlations_disattenuated_prev
        self.contrast_ratios_prev = contrast_ratios_prev

        # Other
        self.timestamps = abs_timestamps - abs_timestamps[0]
        self.framestack = framestack

    def plot(self):
        """
        Plot stability test results: image shift (x and y), correlation, contrast ratio,
        all with respect to first frame and previous frame.
        """
        # Comparisons with first frame
        plt.figure()
        plt.plot(self.timestamps, self.pixel_shifts_first, '.-', label='image-shift (pix)')
        plt.title('Stability - Image shift w.r.t. first frame')
        plt.ylabel('Image shift (pix)')
        plt.xlabel('time (s)')

        plt.figure()
        plt.plot(self.timestamps, self.correlations_first, '.-', label='correlation')
        plt.plot(self.timestamps, self.correlations_disattenuated_first, '.-', label='correlation disattenuated')
        plt.title('Stability - Correlation with first frame')
        plt.xlabel('time (s)')
        plt.legend()

        plt.figure()
        plt.plot(self.timestamps, self.contrast_ratios_first, '.-', label='contrast ratio')
        plt.title('Stability - Contrast ratio with first frame')
        plt.xlabel('time (s)')

        # Comparisons with previous frame
        plt.figure()
        plt.plot(self.timestamps, self.pixel_shifts_prev, '.-', label='image-shift (pix)')
        plt.title('Stability - Image shift w.r.t. previous frame')
        plt.ylabel('Image shift (pix)')
        plt.xlabel('time (s)')

        plt.figure()
        plt.plot(self.timestamps, self.correlations_prev, '.-', label='correlation')
        plt.plot(self.timestamps, self.correlations_disattenuated_prev, '.-', label='correlation disattenuated')
        plt.title('Stability - Correlation with previous frame')
        plt.xlabel('time (s)')
        plt.legend()

        plt.figure()
        plt.plot(self.timestamps, self.contrast_ratios_prev, '.-', label='contrast ratio')
        plt.title('Stability - Contrast ratio with previous frame')
        plt.xlabel('time (s)')

        plt.show()


def measure_setup_stability(frame_source, sleep_time_s, num_of_frames, dark_frame,
                            do_save_frames=False) -> StabilityResult:
    """Test the setup stability by repeatedly reading frames."""
    first_frame = frame_source.read()
    prev_frame = first_frame

    # Initialize arrays - first frame comparisons
    pixel_shifts_first = np.zeros(shape=(num_of_frames, 2))
    correlations_first = np.zeros(shape=(num_of_frames,))
    correlations_disattenuated_first = np.zeros(shape=(num_of_frames,))
    contrast_ratios_first = np.zeros(shape=(num_of_frames,))

    # Initialize arrays - previous frame comparisons
    pixel_shifts_prev = np.zeros(shape=(num_of_frames, 2))
    correlations_prev = np.zeros(shape=(num_of_frames,))
    correlations_disattenuated_prev = np.zeros(shape=(num_of_frames,))
    contrast_ratios_prev = np.zeros(shape=(num_of_frames,))
    abs_timestamps = np.zeros(shape=(num_of_frames,))

    if do_save_frames:
        framestack_shape = first_frame.shape + (num_of_frames,)
        framestack = np.zeros(shape=framestack_shape)
    else:
        framestack = None

    dark_var = dark_frame.var()

    # Start capturing frames
    for n in range(num_of_frames):
        time.sleep(sleep_time_s)
        new_frame = frame_source.read()

        # Compare with first frame
        pixel_shifts_first[n, :] = find_pixel_shift(first_frame, new_frame)
        correlations_first[n] = pearson_correlation(first_frame, new_frame)
        correlations_disattenuated_first[n] = pearson_correlation(first_frame, new_frame, noise_var=dark_var)
        contrast_ratios_first[n] = contrast_enhancement(new_frame, first_frame, dark_frame)

        # Compare with previous frame
        pixel_shifts_prev[n, :] = find_pixel_shift(prev_frame, new_frame)
        correlations_prev[n] = pearson_correlation(prev_frame, new_frame)
        correlations_disattenuated_prev[n] = pearson_correlation(prev_frame, new_frame, noise_var=dark_var)
        contrast_ratios_prev[n] = contrast_enhancement(new_frame, prev_frame, dark_frame)
        abs_timestamps[n] = time.perf_counter()

        # Save frame if requested
        if do_save_frames:
            framestack[:, :, n] = new_frame

        prev_frame = new_frame

    return StabilityResult(pixel_shifts_first=pixel_shifts_first,
                           correlations_first=correlations_first,
                           correlations_disattenuated_first=correlations_disattenuated_first,
                           contrast_ratios_first=contrast_ratios_first,
                           pixel_shifts_prev=pixel_shifts_prev,
                           correlations_prev=correlations_prev,
                           correlations_disattenuated_prev=correlations_disattenuated_prev,
                           contrast_ratios_prev=contrast_ratios_prev,
                           abs_timestamps=abs_timestamps,
                           framestack=framestack)


def measure_modulated_light_dual_phase_stepping(slm: PhaseSLM, feedback: Detector, phase_steps: int, num_blocks: int):
    """
    Measure the ratio of modulated light with the dual phase stepping method.

    Args:
        slm: The SLM that will be used to modulate the light.
        feedback: The Detector providing the feedback.
        phase_steps: Number of phase steps for each group. Must be >=3. Total number of measurements will be
            phase_steps².
        num_blocks: Number of blocks in each dimension of the checkerboard pattern that is used to create the
            modulated groups.

    Returns:
        An estimate of the ratio of modulated light. Independent of any uncorrelated intensity offset.

    Note: In the calculation of the ratio of modulated light, a small epsilon term is added to prevent division by zero.
    """
    assert phase_steps >= 3

    # Initialization
    check_lin = np.arange(num_blocks).reshape(num_blocks, 1)
    block_pattern_p = np.mod(check_lin + check_lin.T, 2)
    block_pattern_q = 1 - block_pattern_p
    measurements = np.zeros((phase_steps, phase_steps))

    # Dual phase stepping
    for p in range(phase_steps):
        phase_p = p * 2 * np.pi / phase_steps
        for q in range(phase_steps):
            phase_q = q * 2 * np.pi / phase_steps
            phase_pattern = block_pattern_p * phase_p + block_pattern_q * phase_q
            slm.set_phases(phase_pattern)
            measurements[p, q] = feedback.read()

    # 2D Fourier transform the modulation measurements
    f = np.fft.fft2(measurements) / phase_steps ** 2

    # Compute fidelity factor due to modulated light
    eps = 1e-6  # Epsilon term to prevent division by zero
    m1_m2_ratio = (np.abs(f[0, 1]) ** 2 + eps) / (np.abs(f[1, 0]) ** 2 + eps)  # Ratio of modulated intensities
    fidelity_modulated = (1 + m1_m2_ratio) / (1 + m1_m2_ratio + np.abs(f[0, 1]) ** 2 / np.abs(f[1, -1]) ** 2)

    return fidelity_modulated


def measure_modulated_light(slm: PhaseSLM, feedback: Detector, phase_steps: int):
    """
    Measure the ratio of modulated light by phase stepping the entire SLM.

    Args:
        slm: The SLM that will be used to modulate the light.
        feedback: The Detector providing the feedback.
        phase_steps: Number of phase steps. Must be >=3.

    Returns:
        fidelity_modulated: An estimate of the ratio of modulated light.

    Note: It is not possible to detect which field is modulated and which is static. This calculation assumes
        the modulated intensity is greater than the non-modulated intensity.
    """
    assert phase_steps >= 3

    # Initialization
    measurements = np.zeros((phase_steps,))

    # Dual phase stepping
    for p in range(phase_steps):
        slm.set_phases(p * 2 * np.pi / phase_steps)
        measurements[p] = feedback.read()

    # 2D Fourier transform the modulation measurements
    f = np.fft.fft(measurements)

    # Compute ratio of modulated light over total
    fidelity_modulated = 0.5 * (1.0 + np.sqrt(np.clip(1.0 - 4.0 * np.abs(f[1] / f[0]) ** 2, 0, None)))

    return fidelity_modulated


class WFSTroubleshootResult:
    """
    Data structure for holding wavefront shaping statistics and additional troubleshooting information.

    Attributes:
        fidelity_non_modulated: The estimated fidelity reduction factor due to the presence of non-modulated light.
        phase_calibration_ratio: A ratio indicating the correctness of the SLM phase response. An incorrect phase
            response produces a value < 1.
        wfs_result (WFSResult): Object containing the analyzed result of running the WFS algorithm.
        frame_signal_std_before: Signal unbiased standard deviation of frame captured before running the WFS algorithm.
        frame_signal_std_after: Signal unbiased standard deviation of frame captured after running the WFS algorithm.
        frame_cnr_before: Unbiased contrast ratio of frame captured before running the WFS algorithm.
        frame_cnr_after: Unbiased contrast ratio of frame captured after running the WFS algorithm.
        frame_contrast_enhancement: The unbiased contrast enhancement due to WFS. A frame taken with a flat wavefront is
            compared to a frame taken with a shaped wavefront. Both frames are taken after the WFS algorithm has run.
        frame_photobleaching_ratio: The signal degradation after WFS, typically caused by photo-bleaching in fluorescent
            experiments. A frame from before and after running the WFS are compared by computing the unbiased contrast
            enhancement factor. If this value is < 1, it means the signal has degraded by this factor.
        frame_repeatability: Pearson correlation of two frames taken before running the WFS algorithm,
            with a flat wavefront. Not compensated for noise.
        dark_frame: Frame taken with the laser blocked, before running the WFS algorithm.
        before_frame: Frame taken with the laser unblocked, before running the WFS algorithm, with a flat wavefront.
        after_frame: Frame taken with the laser unblocked, after running the WFS algorithm, with a flat wavefront.
        shaped_wf_frame: Frame taken with the laser unblocked, after running the WFS algorithm, with a shaped wavefront.
        stability (StabilityResult): Object containing the result of the stability test.
    """

    def __init__(self):
        # Fidelity and WFS metrics
        self.fidelity_non_modulated = None
        self.fidelity_phase_calibration = None
        self.fidelity_decorrelation = None
        self.expected_enhancement = None

        # WFS
        self.wfs_result = None
        self.average_background = None
        self.feedback_before = None
        self.measured_enhancement = None

        # Frame metrics
        self.frame_signal_std_before = None
        self.frame_signal_std_after = None
        self.frame_signal_std_shaped_wf = None
        self.frame_cnr_before = None
        self.frame_cnr_after = None
        self.frame_cnr_shaped_wf = None
        self.frame_contrast_enhancement = None
        self.frame_photobleaching_ratio = None
        self.frame_repeatability = None
        self.stability = None

        # Frames
        self.dark_frame = None
        self.before_frame = None
        self.after_frame = None
        self.shaped_wf_frame = None

        # Other
        self.timestamp = time.time()

    def report(self, do_plots=True):
        """
        Print a report of all results to the console.

        Args:
            do_plots (bool): Plot some results as graphs.
        """
        print(f'\n===========================')
        print(f'{time.ctime(self.timestamp)}\n')
        print(f'=== Feedback metrics ===')
        print(f'number of modes (N): {self.wfs_result.n:.3f}')
        print(f'fidelity_amplitude: {self.wfs_result.fidelity_amplitude.squeeze():.3f}')
        print(f'fidelity_noise: {self.wfs_result.fidelity_noise.squeeze():.3f}')
        print(f'fidelity_non_modulated: {self.fidelity_non_modulated:.3f}')
        print(f'fidelity_phase_calibration: {self.wfs_result.fidelity_calibration.squeeze():.3f}')
        print(f'fidelity_decorrelation: {self.fidelity_decorrelation:.3f}')
        print(f'expected enhancement: {self.expected_enhancement:.3f}')
        print(f'measured enhancement: {self.measured_enhancement:.3f}')
        print(f'')
        print(f'=== Frame metrics ===')
        print(f'signal std, before: {self.frame_signal_std_before:.2f}')
        print(f'signal std, after: {self.frame_signal_std_after:.2f}')
        print(f'signal std, with shaped wavefront: {self.frame_signal_std_shaped_wf:.2f}')
        if self.dark_frame is not None:
            print(f'average offset (dark frame): {self.dark_frame.mean():.2f}')
            print(f'median offset (dark frame): {np.median(self.dark_frame):.2f}')
            print(f'noise std (dark frame): {np.std(self.dark_frame):.2f}')
        print(f'frame repeatability: {self.frame_repeatability:.3f}')
        print(f'contrast to noise ratio before: {self.frame_cnr_before:.3f}')
        print(f'contrast to noise ratio after: {self.frame_cnr_after:.3f}')
        print(f'contrast to noise ratio with shaped wavefront: {self.frame_cnr_shaped_wf:.3f}')
        print(f'contrast enhancement: {self.frame_contrast_enhancement:.3f}')
        print(f'photobleaching ratio: {self.frame_photobleaching_ratio:.3f}')

        if do_plots and self.stability is not None:
            self.stability.plot()

        if (do_plots and self.dark_frame is not None and self.after_frame is not None and
                self.shaped_wf_frame is not None):
            max_value = max(self.after_frame.max(), self.shaped_wf_frame.max())

            # Plot dark frame
            plt.figure()
            plt.imshow(self.dark_frame, vmin=0, vmax=max_value)
            plt.title('Dark frame')
            plt.colorbar()
            plt.xlabel('x (pix)')
            plt.ylabel('y (pix)')
            plt.figure()

            # Plot after frame with flat wf
            plt.imshow(self.after_frame, vmin=0, vmax=max_value)
            plt.title('Frame with flat wavefront')
            plt.colorbar()
            plt.xlabel('x (pix)')
            plt.ylabel('y (pix)')

            # Plot shaped wf frame
            plt.figure()
            plt.imshow(self.shaped_wf_frame, vmin=0, vmax=max_value)
            plt.title('Frame with shaped wavefront')
            plt.colorbar()
            plt.xlabel('x (pix)')
            plt.ylabel('y (pix)')
            plt.show()


def troubleshoot(algorithm, background_feedback: Detector, frame_source: Detector, shutter,
                 do_frame_capture=True, do_long_stability_test=False,
                 stability_sleep_time_s=0.5,
                 stability_num_of_frames=500,
                 stability_do_save_frames=False,
                 measure_non_modulated_phase_steps=16) -> WFSTroubleshootResult:
    """
    Run a series of basic checks to find common sources of error in a WFS experiment.
    Quantifies several types of fidelity reduction.

    Args:
        measure_non_modulated_phase_steps:
        algorithm: Wavefront Shaping algorithm object, e. g. StepwiseSequential.
        background_feedback: Feedback source that determines average background speckle intensity.
        frame_source: Source object for reading frames, e. g. Camera.
        shutter: Device object that can block/unblock light source.
        do_frame_capture: Boolean. If False, skip frame capture before and after running the WFS algorithm.
            Also skips computation of corresponding metrics. Also skips stability test.
        do_long_stability_test: Boolean. If False, skip long stability test where many frames are captured over
            a longer period of time.
        stability_sleep_time_s: Float. Sleep time in seconds in between capturing frames for long stability test.
        stability_num_of_frames: Integer. Number of frames to take in long stability test.
        stability_do_save_frames: Boolean. If True, save all recorded frames.
        measure_non_modulated_phase_steps: Integer. Number of phase steps for determining non-modulated light.


    Returns:
        trouble: WFSTroubleshootResult object containing troubleshoot information.
    """

    # Initialize an empty WFS troubleshoot result
    trouble = WFSTroubleshootResult()

    if do_frame_capture:
        logging.info('Capturing frames before WFS...')

        # Capture frames before WFS
        algorithm.slm.set_phases(0.0)  # Flat wavefront
        shutter.open = False
        trouble.dark_frame = frame_source.read()  # Dark frame
        shutter.open = True
        trouble.before_frame = frame_source.read()  # Before frame (flat wf)
        before_frame_2 = frame_source.read()

        # Frame metrics
        trouble.frame_signal_std_before = signal_std(trouble.before_frame, trouble.dark_frame)
        trouble.frame_cnr_before = cnr(trouble.before_frame, trouble.dark_frame)
        trouble.frame_repeatability = pearson_correlation(trouble.before_frame, before_frame_2)

    if do_long_stability_test and do_frame_capture:
        logging.info('Run long stability test...')

        # Test setup stability
        trouble.stability = measure_setup_stability(
            frame_source=frame_source,
            sleep_time_s=stability_sleep_time_s,
            num_of_frames=stability_num_of_frames,
            dark_frame=trouble.dark_frame,
            do_save_frames=stability_do_save_frames)

    # WFS experiment
    logging.info('Run WFS algorithm...')
    trouble.wfs_result = algorithm.execute()  # Execute WFS algorithm

    # Flat wavefront
    algorithm.slm.set_phases(0.0)
    trouble.average_background = background_feedback.read()

    if do_frame_capture:
        logging.info('Capturing frames after WFS...')
        trouble.after_frame = frame_source.read()  # After frame (flat wf)

    # Shaped wavefront
    algorithm.slm.set_phases(-np.angle(trouble.wfs_result.t))
    trouble.feedback_shaped_wf = algorithm.feedback.read()

    trouble.measured_enhancement = trouble.feedback_shaped_wf / trouble.average_background

    if do_frame_capture:
        trouble.shaped_wf_frame = frame_source.read()  # Shaped wavefront frame

        # Frame metrics
        logging.info('Compute frame metrics...')
        trouble.frame_signal_std_after = signal_std(trouble.after_frame, trouble.dark_frame)
        trouble.frame_signal_std_shaped_wf = signal_std(trouble.shaped_wf_frame, trouble.dark_frame)
        trouble.frame_cnr_after = cnr(trouble.after_frame, trouble.dark_frame)  # Frame CNR after
        trouble.frame_cnr_shaped_wf = cnr(trouble.shaped_wf_frame, trouble.dark_frame)  # Frame CNR shaped wf
        trouble.frame_contrast_enhancement = \
            contrast_enhancement(trouble.shaped_wf_frame, trouble.after_frame, trouble.dark_frame)
        trouble.frame_photobleaching_ratio = \
            contrast_enhancement(trouble.after_frame, trouble.before_frame, trouble.dark_frame)
        trouble.fidelity_decorrelation = \
            pearson_correlation(trouble.before_frame, trouble.after_frame, noise_var=trouble.dark_frame.var())

    trouble.fidelity_non_modulated = \
        measure_modulated_light(slm=algorithm.slm, feedback=algorithm.feedback,
                                phase_steps=measure_non_modulated_phase_steps)

    trouble.expected_enhancement = np.squeeze(
        trouble.wfs_result.n * trouble.wfs_result.fidelity_amplitude * trouble.wfs_result.fidelity_noise
        * trouble.fidelity_non_modulated * trouble.wfs_result.fidelity_calibration * trouble.fidelity_decorrelation)

    # Analyze the WFS result
    logging.info('Analyze WFS result...')
    return trouble



from . import microscope
from . import mockdevices
from . import slm
from . import transmission

from .microscope import Microscope
from .mockdevices import XYStage, StaticSource, Camera, ADCProcessor, Shutter, NoiseSource
from .slm import SLM, PhaseToField
from .transmission import SimulatedWFS



from typing import Optional

import numpy as np

from .slm import SLM, Processor


class SimulatedWFS(Processor):
    """A simple simulation of a wavefront shaping experiment.

    Simulates a configuration where both the SLM and the aberrations are placed in the back pupil
    of a microscope objective, and a point detector is placed in the center of the focal plane.

    The simulation computes (Σ A·exp(i·(aberrations-slm)))², which is the 0,0 component of the Fourier transform
    of the field in the pupil plane (including aberrations and the correction by the SLM).

    For a more advanced (but slower) simulation, use `Microscope`
    """

    def __init__(self, *, t: Optional[np.ndarray] = None, aberrations: Optional[np.ndarray] = None, slm=None,
                 multi_threaded=True):
        """
        Initializes the optical system with specified aberrations and optionally a Gaussian beam profile.

        This constructor sets up an optical system by specifying the aberrations across the SLM (Spatial Light
        Modulator) and, optionally, by defining a Gaussian beam profile. The aberrations array defines phase shifts at
        each point of the SLM. If a beam profile waist is provided, the input electric field is shaped into a
        Gaussian beam.

        Args:
            t: Transmission matrix.
            aberrations: An array containing the aberrations in radians. Can be used instead of a transmission matrix,
                equivalent to specifying t = np.exp(1j * aberrations) / (aberrations.shape[0] * aberrations.shape[1]).
            slm:
            multi_threaded (bool, optional): If True, the simulation will use multiple threads to compute the
                intensity in the focus. If False, the simulation will use a single thread. Defaults to True.

        The constructor creates a MockSLM instance based on the shape of the aberrations, calculates the electric
        field at the SLM considering the aberrations and optionally the Gaussian beam profile, and initializes the
        system with these parameters.
        """

        # transmission matrix (normalized so that the maximum transmission is 1)
        self._t = t if t is not None else np.exp(1.0j * aberrations) / (aberrations.shape[0] * aberrations.shape[1])
        self.slm = slm if slm is not None else SLM(self._t.shape[0:2])

        super().__init__(self.slm.field, multi_threaded=multi_threaded)

    def _fetch(self, incident_field):  # noqa
        """
        Computes the intensity in the focus by applying phase corrections to the input electric field.

        This method adjusts the phase of the input electric field using the SLM (Spatial Light Modulator) phases,
        computes the tensor product with the electric field at the SLM, and then calculates the intensity of the
        resulting field.

        Args:
            incident_field: The incident field (complex)

        Returns:
            np.ndarray: A numpy array containing the calculated intensity in the focus.

        """
        field = np.tensordot(incident_field, self._t, 2)
        return np.abs(field) ** 2

    @property
    def data_shape(self):
        return self._t.shape[2:]



from typing import Sequence, Optional

import cv2
import numpy as np
from astropy.units import Quantity

from ..core import Processor, Detector
from ..utilities import project, Transform
from ..utilities.patterns import disk, gaussian


class Roi:
    """
    Represents a Region of Interest (ROI) to compute a (weighted) average over.

    This class defines an ROI with specified properties such as coordinates,
    radius, mask type, and parameters specific to the mask type.
    """

    def __init__(self, pos, radius=0.1, mask_type: str = 'disk', waist=None,
                 source_shape=None):
        """
        Initialize the Roi object.

        Args:
            pos (int, int): y,x coordinates of the center of the ROI, measured in pixels from the top-left corner.
                when omitted, the default value of source_shape // 2 is used.
                note: non-integer positions for the ROI are currently not supported.
            radius (float): Radius of the ROI. Default is 0.1.
            mask_type: Type of the mask.
                Options are 'disk' (default), 'gaussian', or 'square'.
            waist (float): Defines the width of the Gaussian distribution in pixels.
                Default is 0.5 * radius.
            source_shape (int, int): Shape of the source image.
                Used to compute a default value for `pos`, and to check if the ROI is fully inside the image.
        """
        if pos is None:
            pos = (source_shape[0] // 2, source_shape[1] // 2)
        if round(pos[0] - radius) < 0 or round(pos[1] - radius) < 0 or source_shape is not None and (
                round(pos[0] + radius) >= source_shape[0] or
                round(pos[1] + radius) >= source_shape[1]):
            raise ValueError('ROI does not fit inside source image')

        self._pos = pos
        self._radius = radius
        self._mask_type = mask_type
        self._waist = waist if waist is not None else radius * 0.5
        self._mask = None
        self._mask_sum = 0.0

    @property
    def pos(self) -> tuple[int, int]:
        return self._pos

    @pos.setter
    def pos(self, value: tuple[int, int]):
        self._pos = value
        self._mask = None  # need to re-compute mask

    @property
    def x(self) -> int:
        """x-coordinate of the center of the ROI, relative to the center of the image."""
        return self.pos[1]

    @x.setter
    def x(self, value: int):
        self.pos = (self.pos[0], int(value))

    @property
    def y(self) -> int:
        """y-coordinate of the center of the ROI, relative to the center of the image."""
        return self.pos[0]

    @y.setter
    def y(self, value: int):
        self.pos = (int(value), self.pos[1])

    @property
    def radius(self) -> float:
        return self._radius

    @radius.setter
    def radius(self, value: float):
        self._radius = float(value)
        self._mask = None  # need to re-compute mask

    @property
    def waist(self) -> float:
        return self._waist

    @waist.setter
    def waist(self, value: float):
        self._waist = float(value)
        self._mask = None  # need to re-compute mask

    @property
    def mask_type(self) -> str:
        return self._mask_type

    @waist.setter
    def waist(self, value: str):
        if value not in ['disk', 'gaussian', 'square']:
            raise ValueError("mask_type must be 'disk', 'gaussian', or 'square'")
        self._mask_type = value
        self._mask = None  # need to re-compute mask

    def apply(self, image: np.ndarray, order: float = 1.0):
        """
        Applies the mask to the frame data by computing the weighted average.
        Optionally, the image data can be raised to a power before the mask is applied.
        This is useful for simulating multi-photon excitation, or for computing
        weighted contrast over the mask.

        Args:
            image (np.ndarray): The source image data.
            order (float): The order of the mask. Default is 1.0.
        """

        # if any of the variables changed, we need to re-compute the mask
        if self._mask is None:
            # clip the radius so that it corresponds to at least 1 pixel
            r = np.maximum(self._radius, 0.1)

            # for circular masks, always use an odd number of pixels so that we have a clearly
            # defined center.
            # for square masks, instead use the actual size
            if self.mask_type == 'disk':
                d = round(self._radius) * 2 + 1
                self._mask = disk(d, r)
            elif self.mask_type == 'gaussian':
                d = round(self._radius) * 2 + 1
                self._mask = gaussian(d, self._waist)
            else:  # square
                d = round(self._radius * 2.0)
                self._mask = np.ones((d, d))

            self._mask_sum = np.sum(self._mask)

        image_start = np.array(self.pos) - int(0.5 * self._mask.shape[0] - 0.5)
        image_cropped = image[
                        image_start[0]:image_start[0] + self._mask.shape[0],
                        image_start[1]:image_start[1] + self._mask.shape[1]]

        if image_cropped.shape != self._mask.shape:
            raise ValueError(
                f"ROI is larger than the possible area. ROI shape: {self._mask.shape}, "
                + f"Cropped image shape: {image_cropped.shape}")

        if order != 1.0:
            image_cropped = np.power(image_cropped, order)

        return np.sum(image_cropped * self._mask) / self._mask_sum


class MultipleRoi(Processor):
    """
    Processor that averages signals over multiple regions of interest (ROIs).
    """

    def __init__(self, source, rois: Sequence[Roi], multi_threaded: bool = True):
        """
        Initialize the MultipleRoi processor with a source and multiple ROIs.
        Note: changing parameters of the ROIs between triggering and fetching causes a race condition.
        Args:
            source (Detector): Source detector object to process the data from.
            rois (Sequence[Roi]): Sequence of Roi objects defining the regions of interest.
        """
        self._rois = np.array(rois)
        self._source = source
        super().__init__(source, multi_threaded=multi_threaded)

    def _fetch(self, image: np.ndarray) -> np.ndarray:  # noqa
        """
        Fetches and processes the data for each ROI from the image.

        This method crops the image according to each ROI's position and
        calculates the average value within the ROI. If an ROI is larger than
        the possible area of the image, a ValueError is raised.

        Args:
            image (np.ndarray): The source image data.

        Returns:
            np.ndarray: Array containing the processed data for each ROI.
        """

        def apply_mask(mask: Roi):
            return mask.apply(image)

        return np.vectorize(apply_mask)(self._rois)

    @property
    def data_shape(self):
        return self._rois.shape

    @property
    def pixel_size(self) -> None:
        """Returns None, since the elements in the output of the MultipleRoi processor do not have a physical size."""
        return None


class SingleRoi(MultipleRoi):
    def __init__(self, source, pos=None, radius=0.1, mask_type: str = 'disk', waist=0.5,
                 multi_threaded: bool = True):
        """
        Processor that averages a signal over a single region of interest (ROI).

        Args:
            source (Detector): Source detector object to process the data from.
            pos (int, int): y,x coordinates of the center of the ROI, measured in pixels from the top-left corner.
                when omitted, the default value of source.data_shape // 2 is used.
                note: non-integer positions for the ROI are currently not supported.
            radius (float): Radius of the ROI. Default is 0.1.
            mask_type: Type of the mask. Options are 'disk', 'gaussian', or 'square'. Default is 'disk'.
            waist (float): Defines the width of the Gaussian distribution. Default is 0.5.
        """
        single_roi = Roi(pos, radius, mask_type, waist, source.data_shape)
        rois = np.array([single_roi]).reshape(())
        super().__init__(source, rois=rois, multi_threaded=multi_threaded)
        self.__dict__.update(single_roi.__dict__)


class CropProcessor(Processor):
    """Processor to crop data from the source to some region of interest.

    Works on any number of dimensions.
    If the cropped area extends beyond the size of the source data,
    the data is padded with 'padding_value'
    """

    def __init__(self, source: Detector, shape: Optional[Sequence[int]] = None,
                 pos: Optional[Sequence[int]] = None, padding_value=0.0, multi_threaded: bool = False):
        """

        Args:
            source (object): The data source to process.
            shape (tuple): Size of the cropped region (this is data_shape property)
                default is None: use the full size of the source.
                may be a tuple holding one or more None values.
                These values are then replaced by the size of the source in that dimension.
            pos (tuple): Coordinates of the start of the cropped region.
                For 2-D data, this is the top-left corner.
            padding_value (float): Value to use if the cropped area extends beyond the original data.
        """
        super().__init__(source, multi_threaded=multi_threaded)
        self._data_shape = tuple(shape) if shape is not None else source.data_shape
        self._pos = np.array(pos) if pos is not None else np.zeros((len(self.data_shape),), dtype=int)
        self._padding_value = padding_value

    @property
    def pos(self) -> tuple:
        """Start ('top-left' corner) of the cropped region."""
        return tuple(self._pos)

    @pos.setter
    def pos(self, value):
        self._pos = np.array(value, ndmin=1)

    @property
    def data_shape(self) -> tuple:
        return self._data_shape

    @data_shape.setter
    def data_shape(self, value):
        self._data_shape = tuple(np.array(value, ndmin=1))

    def _fetch(self, image: np.ndarray) -> np.ndarray:  # noqa
        """
        Args:
            image (ndarray): source image

        Returns: the out array containing the cropped image.

        """
        src_start = np.maximum(self._pos, 0).astype('int32')
        src_end = np.minimum(self._pos + self._data_shape, image.shape).astype('int32')
        dst_start = np.maximum(-self._pos, 0).astype('int32')
        dst_end = dst_start + src_end - src_start
        src_select = tuple(
            slice(start, end) for (start, end) in zip(src_start, src_end))
        src = image.__getitem__(src_select)
        if any(dst_start != 0) or any(dst_end != self._data_shape):
            dst = np.zeros(self._data_shape) + self._padding_value
            dst_select = tuple(slice(start, end) for (start, end) in zip(dst_start, dst_end))
            dst.__setitem__(dst_select, src)
        else:
            dst = src

        return dst


def select_roi(source: Detector, mask_type: str):
    """
    Opens a window that allows the user to select a region of interest.
    """
    if mask_type not in ['disk', 'gaussian', 'square']:
        raise ValueError("mask_type must be 'disk', 'gaussian', or 'square'")

    image = cv2.normalize(source.read(), None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)
    title = "Select ROI and press c to continue or ESC to cancel"
    cv2.namedWindow(title)
    cv2.imshow(title, image)

    roi_start = np.array((0, 0))
    roi_size = 0.0

    def mouse_callback(event, x, y, flags, _param):
        nonlocal roi_start, roi_size, image
        if event == cv2.EVENT_LBUTTONDOWN:  # mouse down: select start
            roi_start = np.array((x, y))

        elif event == cv2.EVENT_MOUSEMOVE and cv2.EVENT_FLAG_LBUTTON & flags:
            roi_size = np.minimum(x - roi_start[0], y - roi_start[1])
            rect_image = image.copy()
            if mask_type == 'square':
                cv2.rectangle(rect_image, roi_start, roi_start + roi_size, (0.0, 0.0, 255.0), 2)
            else:
                cv2.circle(rect_image, roi_start + roi_size // 2, abs(roi_size) // 2, (0.0, 0.0, 255.0), 2)
            cv2.imshow(title, rect_image)

    cv2.setMouseCallback(title, mouse_callback)

    while True:
        key = cv2.waitKey(1) & 0xFF
        if key == ord("c"):
            if roi_size is not None:
                cv2.destroyWindow(title)
                if roi_size < 0.0:
                    roi_start = roi_start + roi_size
                    roi_size = -roi_size
                return Roi(pos=(roi_start[1], roi_start[0]), radius=0.5 * roi_size)

        elif key == 27:
            cv2.destroyWindow(title)
            return None


class TransformProcessor(Processor):
    """
    Performs a 2-D transform of the input data (including shifting, padding, cropping, resampling).

    By default, the output shape and pixel_size are the same as the input shape.
    If desired, explicit values can be provided for the output shape and pixel_size. The unit of the pixel_size
    should match the unit of the input data after applying the transform.
    """

    def __init__(self, source: Detector,
                 transform: Transform = None,
                 data_shape: Optional[Sequence[int]] = None,
                 pixel_size: Optional[Quantity] = None,
                 multi_threaded: bool = True):
        """

        Args:
            transform: Transform object that describes the transformation from the source to the target image
            data_shape: Shape of the output. If omitted, the shape of the input data is used.
            multi_threaded: Whether to perform processing in a worker thread.
            """
        if (data_shape is not None and len(data_shape) != 2) or len(source.data_shape) != 2:
            raise ValueError("TransformProcessor only supports 2-D data")
        if transform is None:
            transform = Transform()

        # check if input and output pixel sizes are compatible
        dst_unit = transform.destination_unit(source.pixel_size.unit)
        if pixel_size is not None and not pixel_size.unit.is_equivalent(dst_unit):
            raise ValueError("Pixel size unit does not match the unit of the transformed data")
        if pixel_size is None and not source.pixel_size.unit.is_equivalent(dst_unit):
            raise ValueError("The transform changes the unit of the coordinates."
                             " An output pixel_size must be provided.")

        self.transform = transform
        super().__init__(source, multi_threaded=multi_threaded)
        self._pixel_size = pixel_size
        self._data_shape = data_shape

    @property
    def data_shape(self):
        return self._data_shape if self._data_shape is not None else super().data_shape

    @property
    def pixel_size(self) -> Optional[Quantity]:
        return self._pixel_size if self._pixel_size is not None else super().pixel_size

    def _fetch(self, source: np.ndarray) -> np.ndarray:  # noqa
        """
        Args:
            source (Detector): A Detector object as described in openwfs.core.Detector

        Returns: ndarray that has been transformed
        TODO: Fix and add test, or remove
        """
        return project(source, transform=self.transform, out_shape=self.data_shape,
                       out_extent=self.extent)



from astropy import units as u
from matplotlib import pyplot as plt

from .core import Detector
from .utilities import get_extent


def grab_and_show(cam: Detector, axis=None):
    return imshow(cam.read(), axis=axis)


def imshow(data, axis=None):
    extent = get_extent(data)
    e0 = scale_prefix(extent[0])
    e1 = scale_prefix(extent[1])
    if axis is None:
        plt.imshow(data, extent=(0.0, e1.value, 0.0, e0.value), cmap='gray')
        plt.colorbar()
        axis = plt.gca()
    else:
        axis.imshow(data, extent=(0.0, e1.value, 0.0, e0.value), cmap='gray')
    plt.ylabel(e0.unit.to_string())
    plt.xlabel(e1.unit.to_string())
    plt.show(block=False)
    plt.pause(0.1)
    return axis


def scale_prefix(value: u.Quantity) -> u.Quantity:
    """Scale a quantity to the most appropriate prefix unit."""
    if value.unit.physical_type == 'length':
        if value < 100 * u.nm:
            return value.to(u.nm)
        if value < 100 * u.um:
            return value.to(u.um)
        if value < 100 * u.mm:
            return value.to(u.mm)
        else:
            return value.to(u.m)
    elif value.unit.physical_type == 'time':
        if value < 100 * u.ns:
            return value.to(u.ns)
        if value < 100 * u.us:
            return value.to(u.us)
        if value < 100 * u.ms:
            return value.to(u.ms)
        else:
            return value.to(u.s)
    else:
        return value



import warnings
from typing import Sequence

import numpy as np
from numpy.typing import ArrayLike

from .context import Context

try:
    import OpenGL.GL as GL
    from OpenGL.GL import glGenBuffers, glBindBuffer, glBufferData, glDeleteBuffers, glEnable, glBlendFunc, \
        glBlendEquation, glDisable, glUseProgram, glBindVertexBuffer, glDrawElements, glGenFramebuffers, \
        glBindFramebuffer, glFramebufferTexture2D, glCheckFramebufferStatus, glDeleteFramebuffers, \
        glEnableVertexAttribArray, glVertexAttribFormat, glVertexAttribBinding, glEnableVertexAttribArray, \
        glPrimitiveRestartIndex, glActiveTexture, glBindTexture, glGenVertexArrays, glBindVertexArray

    from OpenGL.GL import shaders
except AttributeError:
    warnings.warn("OpenGL not found, SLM will not work")
from .geometry import rectangle, Geometry
from .shaders import default_vertex_shader, default_fragment_shader, \
    post_process_fragment_shader, post_process_vertex_shader
from .texture import Texture
from ...core import PhaseSLM


class Patch(PhaseSLM):
    _PHASES_TEXTURE = 0  # indices of the phases texture in the _texture array

    def __init__(self, slm, geometry=None, vertex_shader=default_vertex_shader,
                 fragment_shader=default_fragment_shader):
        """
        Constructs a new patch (a shape) that can be drawn on the screen.
        By default, the patch is a square with 'radius' 1.0 (width and height 2.0) centered at 0.0, 0.0
        To specify a different geometry, provide either:
          - a 2-D array of vertices, such as produced by the 'geometry' module. The vertices are interpreted as points
            on a (possibly deformed) grid.
          - a tuple containing a 1-D array of vertices and a 1-D array of indices. The indices indicate how the
            vertices are connected into triangles (see Geometry object for details).
          - an existing Geometry object. It is possible to attach the same Geometry object to multiple patches.
            Note, however, that Geometry objects cannot be shared between different SLMs.
        """
        self._vertices = None
        self._indices = None
        self._index_count = 0
        self.additive_blend = True
        self.enabled = True
        self.context = Context(slm)

        # construct vertex shader, fragment shader and program
        with self.context:
            vs = shaders.compileShader(vertex_shader, GL.GL_VERTEX_SHADER)
            fs = shaders.compileShader(fragment_shader, GL.GL_FRAGMENT_SHADER)
            self._program = shaders.compileProgram(vs, fs)
            self._textures = [Texture(self.context)]

        self.geometry = rectangle(2.0) if geometry is None else geometry
        super().__init__()

    def __del__(self):
        self._delete_buffers()

    def _draw(self):
        """Never call directly, this is called from slm.update()"""
        if not self.enabled:
            return

        glUseProgram(self._program)

        if self.additive_blend:
            glEnable(GL.GL_BLEND)
            glBlendFunc(GL.GL_ONE, GL.GL_ONE)  # (1 * rgb, 1 * alpha)
            glBlendEquation(GL.GL_FUNC_ADD)
        else:
            glDisable(GL.GL_BLEND)

        for idx, texture in enumerate(self._textures):
            # activate texture as texture unit idx
            texture._bind(idx)  # noqa: ok to use _bind in friend class

        # perform the actual drawing
        glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, self._indices)
        glBindVertexBuffer(0, self._vertices, 0, 16)
        glDrawElements(GL.GL_TRIANGLE_STRIP, self._index_count, GL.GL_UNSIGNED_SHORT, None)

    def set_phases(self, values: ArrayLike, update=True):
        """
        Args:
            values(ArrayLike): 1-D or 2-D array holding phase values to display on the SLM.
                Phases are in radians, and stored as float32. There is no need to wrap the phase to a 0-2pi range.
            update(bool): when True, the SLM in which this patch is contained is updated immediately.
                When False, the SLM is not updated, and the
                caller is responsible for calling slm.update() to update the SLM.
        """
        self._textures[Patch._PHASES_TEXTURE].set_data(values)
        if update:
            self.update()

    def update(self):
        self.context.slm.update()

    def _delete_buffers(self):
        with self.context as slm:
            if slm:
                glDeleteBuffers(2, [self._vertices, self._indices])

    @property
    def geometry(self):
        """Vertices that define the shape of the patch on the screen. Currently, this should be a NxMx4 numpy array
        of float32 values. Each 4 values define a vector: x,y position and tx, ty texture coordinate. See geometry.py
        for examples of geometry specifications. The vertices are drawn as a NxM 'rectangular' grid of quadrilaterals.
        """
        return self._geometry

    @geometry.setter
    def geometry(self, value: Geometry):
        if not isinstance(value, Geometry):
            raise ValueError("Geometry should be a Geometry object")

        # store the data on the GPU
        with self.context:
            self._geometry = value
            (self._vertices, self._indices) = glGenBuffers(2)
            self._index_count = value.indices.size
            glBindBuffer(GL.GL_ARRAY_BUFFER, self._vertices)
            glBufferData(GL.GL_ARRAY_BUFFER, value.vertices.size * 4, value.vertices, GL.GL_DYNAMIC_DRAW)
            glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, self._indices)
            glBufferData(GL.GL_ELEMENT_ARRAY_BUFFER, value.indices.size * 2, value.indices, GL.GL_DYNAMIC_DRAW)


class FrameBufferPatch(Patch):
    """Special patch that represents the frame buffer. All patches are first rendered to the frame buffer,
    and this buffer is rendered to the screen through a  final post-processing step that does the phase wrapping and
    implements the software lookup table."""

    _LUT_TEXTURE = 1
    _textures: list[Texture]

    def __init__(self, slm, lookup_table: Sequence[int]):
        super().__init__(slm, fragment_shader=post_process_fragment_shader,
                         vertex_shader=post_process_vertex_shader)
        # Create a frame buffer object to render to. The frame buffer holds a texture that is the same size as the
        # window. All patches are first rendered to this texture. The texture
        # is then processed as a whole (applying the software lookup table) and displayed on the screen.
        self._frame_buffer = glGenFramebuffers(1)

        self.set_phases(np.zeros(self.context.slm.shape, dtype=np.float32), update=False)
        glBindFramebuffer(GL.GL_FRAMEBUFFER, self._frame_buffer)
        glFramebufferTexture2D(GL.GL_FRAMEBUFFER, GL.GL_COLOR_ATTACHMENT0, GL.GL_TEXTURE_2D,
                               self._textures[Patch._PHASES_TEXTURE].handle, 0)
        if glCheckFramebufferStatus(GL.GL_FRAMEBUFFER) != GL.GL_FRAMEBUFFER_COMPLETE:
            raise Exception("Could not construct frame buffer")
        glBindFramebuffer(GL.GL_FRAMEBUFFER, 0)

        self._textures.append(Texture(self.context, GL.GL_TEXTURE_1D))  # create texture for lookup table
        self._lookup_table = None
        self.lookup_table = lookup_table
        self.additive_blend = False

    def __del__(self):
        with self.context as slm:
            if slm:
                glDeleteFramebuffers(1, [self._frame_buffer])

    @property
    def lookup_table(self):
        """1-D array """
        return self._lookup_table

    @lookup_table.setter
    def lookup_table(self, value):
        self._lookup_table = np.array(value)
        self._textures[FrameBufferPatch._LUT_TEXTURE].set_data(self._lookup_table / 255)

    def get_pixels(self):
        data = self._textures[FrameBufferPatch._PHASES_TEXTURE].get_data()

        # flip data upside down, because the OpenGL convention is to have the origin at the bottom left,
        # but we want it at the top left (like in numpy)
        return data[::-1, :]


class VertexArray:
    # A VertexArray informs OpenGL about the format of the vertex data we will use.
    # Each vertex contains four float32 components:
    # x, y coordinate for vertex position. Will be transformed by the transform matrix.
    # tx, ty texture coordinates, range from 0.0, 0.0 to 1.0, 1.0 to cover the full texture
    #
    # To inform OpenGL about this format, we create vertex an array object and store format properties.
    # The elements of the vertex are available to a vertex shader as vec2 position (location 0)
    # and vec2 texture_coordinates (location 1), see vertex shader in Patch for an example.
    # All this information is bound to a binding index before use by calling glBindVertexBuffer,
    # which is done when a vertex buffer is created (see Patch).
    #
    # Since we have a fixed vertex format, we only need to bind the VertexArray once, and not bother with
    # updating, binding, or even deleting it
    def __init__(self):
        self._vertex_array = glGenVertexArrays(1)  # no need to destroy explicitly, destroyed when window is destroyed
        glBindVertexArray(self._vertex_array)
        glEnableVertexAttribArray(0)
        glEnableVertexAttribArray(1)
        glVertexAttribFormat(0, 2, GL.GL_FLOAT, GL.GL_FALSE, 0)  # first two float32 are screen coordinates
        glVertexAttribFormat(1, 2, GL.GL_FLOAT, GL.GL_FALSE, 8)  # second two are texture coordinates
        glVertexAttribBinding(0, 0)  # use binding index 0 for both attributes
        glVertexAttribBinding(1, 0)  # the attribute format can now be used with glBindVertexBuffer

        # enable primitive restart, so that we can draw multiple triangle strips with a single draw call
        glEnable(GL.GL_PRIMITIVE_RESTART)
        glPrimitiveRestartIndex(0xFFFF)  # this is the index we use to separate individual triangle strips



from . import patterns
from . import utilities
from .patterns import coordinate_range, disk, gaussian, tilt
from .utilities import ExtentType, CoordinateType, unitless, get_pixel_size, \
    set_pixel_size, Transform, project, place, set_extent, get_extent



from dataclasses import dataclass
from typing import Union, Sequence, Optional

import cv2
import numpy as np
from astropy import units as u
from astropy.units import Quantity
from numpy.typing import ArrayLike

# A coordinate is a sequence of two floats with an optional unit attached
CoordinateType = Union[Sequence[float], np.ndarray, Quantity]

# A transform is a 2x2 array with optional units attached
TransformType = Union[np.ndarray, Quantity, Sequence[Sequence[float]]]

# An extent is like a coordinate type.
# If it has a single element, this element is automatically reproduced along all dimensions
ExtentType = Union[float, CoordinateType]


def unitless(data: ArrayLike) -> np.ndarray:
    """
    Converts an object to a numpy array.
    Especially useful to remove the scaling from a unitless Quantity.

    Args:
        data: The input data.
        If `data` is a Quantity, it is converted to a (unitless) numpy array.
        All other data types are just returned as is.

    Returns:
        ArrayLike: unitless numpy array, or the input data if it is not a Quantity.

    Raises:
        UnitConversionError: If the data is a Quantity with a unit

    Note:
        Do NOT use `np.array(data)` to convert a Quantity to a numpy array,
        because this will drop the unit prefix.
        For example, `np.array(1 * u.s / u.ms) == 1`.
        Whereas `unitless(1 * u.s / u.ms) == 1000` gives the correct answer.

    Usage:
    >>> data = np.array([1.0, 2.0, 3.0]) * u.m
    >>> unitless_data = unitless(data)
    """
    if isinstance(data, Quantity):
        return data.to_value(u.dimensionless_unscaled)
    else:
        return np.array(data)


@dataclass
class Transform:
    """Represents a transformation from one 2-d coordinate system to another.

    Transform objects are used to specify any combination of shift, (anisotropic) scaling, rotation and shear.
    The transform matrix, as well as the source and destination origins, can be specified with astropy units attached.

    Note that a Transform object does not contain any information about the extent and pixel size
    of the source or destination image, it only specifies the coordinate transformation itself.
    When a Transform object is used to transform image data (see :func:`project` and :func:`place`),
    the pixel_size information should be specified separately.
    """

    """
    Args:
        transform:
            2x2 transformation matrix that transforms (y,x)
            coordinates in the source image to (y,x) coordinates in the destination image.
            This matrix may include astropy units,
            for instance when converting from normalized SLM coordinates to physical coordinates in micrometers.
            Note:
                the units of the transform must match the units of the pixel_size of the destination
                divided by the pixel size of the source.
        source_origin:
            (y,x) coordinate of the origin in the source image, relative to the center of the image.
            By default, the center of the source image is taken as the origin of the transform,
            meaning that that point is mapped onto the destination_origin.
            Note:
                the units of source_origin must match the units of the pixel_size of the source.
        destination_origin:
            (y,x) coordinate of the origin in the destination image, relative to the center of the image.
            By default, the center of the destination image is taken as the origin of the transform,
            meaning that the source_origin is mapped to that point.
            Note:
                the units of destination_origin must match the units of the pixel_size of the source.


    """

    def __init__(self, transform: Optional[TransformType] = None,
                 source_origin: Optional[CoordinateType] = None,
                 destination_origin: Optional[CoordinateType] = None):

        self.transform = Quantity(transform if transform is not None else np.eye(2))
        self.source_origin = Quantity(source_origin) if source_origin is not None else None
        self.destination_origin = Quantity(destination_origin) if destination_origin is not None else None

        if source_origin is not None:
            self.destination_unit(self.source_origin.unit)  # check if the units are consistent

    def destination_unit(self, src_unit: u.Unit) -> u.Unit:
        """Computes the unit of the output of the transformation, given the unit of the input.

        Raises:
            ValueError: If src_unit does not match the unit of the source_origin (if specified) or
                if dst_unit does not match the unit of the destination_origin (if specified).
        """
        if self.source_origin is not None and not self.source_origin.unit.is_equivalent(src_unit):
            raise ValueError("src_unit must match the units of source_origin.")

        dst_unit = (self.transform[0, 0] * src_unit).unit
        if self.destination_origin is not None and not self.destination_origin.unit.is_equivalent(dst_unit):
            raise ValueError("dst_unit must match the units of destination_origin.")

        return dst_unit

    def cv2_matrix(self,
                   source_shape: Sequence[int],
                   source_pixel_size: CoordinateType,
                   destination_shape: Sequence[int],
                   destination_pixel_size: CoordinateType) -> np.ndarray:
        """Returns the transformation matrix in the format used by cv2.warpAffine."""

        # correct the origin. OpenCV uses the _center_ of the top-left corner as the origin
        # and openwfs uses the center of the image as the origin, so we need to shift the origin
        # by half the image size minus half a pixel.
        if min(source_shape) < 1 or min(destination_shape) < 1:
            raise ValueError("Image size must be positive")

        source_origin = 0.5 * (np.array(source_shape) - 1.0) * source_pixel_size
        if self.source_origin is not None:
            source_origin += self.source_origin

        destination_origin = 0.5 * (np.array(destination_shape) - 1.0) * destination_pixel_size
        if self.destination_origin is not None:
            destination_origin += self.destination_origin

        centered_transform = Transform(transform=self.transform,
                                       source_origin=source_origin,
                                       destination_origin=destination_origin)

        # then convert the transform to a matrix, using the specified pixel sizes
        transform_matrix = centered_transform.to_matrix(source_pixel_size=source_pixel_size,
                                                        destination_pixel_size=destination_pixel_size)

        # finally, convert the matrix to the format used by cv2.warpAffine by swapping x and y columns and rows
        transform_matrix = transform_matrix[[1, 0], :]
        transform_matrix = transform_matrix[:, [1, 0, 2]]
        return transform_matrix

    def to_matrix(self, source_pixel_size: CoordinateType, destination_pixel_size: CoordinateType) -> np.ndarray:
        matrix = np.zeros((2, 3))
        matrix[0:2, 0:2] = unitless(self.transform * source_pixel_size / destination_pixel_size)
        if self.destination_origin is not None:
            matrix[0:2, 2] = unitless(self.destination_origin / destination_pixel_size)
        if self.source_origin is not None:
            matrix[0:2, 2] -= unitless((self.transform @ self.source_origin) / destination_pixel_size)
        return matrix

    def opencl_matrix(self) -> np.ndarray:
        # compute the transform for points (0,1), (1,0), and (0, 0)
        matrix = self.to_matrix((1.0, 1.0), (1.0, 1.0))

        # subtract the offset (transform of 0,0) from the first two points
        # to construct the homogeneous transformation matrix
        # convert to opencl format: swap x and y columns (note: the rows were
        # already swapped in the construction of t2), and flip the sign of the y-axis.
        transform = np.eye(3, 4, dtype='float32', order='C')
        transform[0, 0:3] = matrix[1, [1, 0, 2],]
        transform[1, 0:3] = -matrix[0, [1, 0, 2],]
        return transform

    @staticmethod
    def zoom(scale: float):
        """Returns a transform that just zooms in the image."""
        return Transform(transform=np.array(((scale, 0.0), (0.0, scale))))

    def __matmul__(self, other):
        """The matrix multiplication operator is used to compose transformations,
        and to apply a transformation to a vector."""
        if isinstance(other, Transform):
            return self.compose(other)
        else:
            return self.apply(other)

    def apply(self, vector: CoordinateType) -> CoordinateType:
        """Applies the transformation to a column vector.

         If `vector` is a 2-D array, applies the transformation to each column of `vector` individually."""
        if self.source_origin is not None:
            vector = vector - self.source_origin
        vector = self.transform @ vector
        if self.destination_origin is not None:
            vector = vector + self.destination_origin
        return vector

    def inverse(self):
        """Compute the inverse transformation,
        such that the composition of the transformation and its inverse is the identity."""

        # invert the transform matrix
        if self.transform is not None:
            transform = np.linalg.inv(self.transform)
        else:
            transform = None

        # swap source and destination origins
        return Transform(transform, source_origin=self.destination_origin, destination_origin=self.source_origin)

    def compose(self, other: 'Transform'):
        """Compose two transformations.

        Args:
            other (Transform): the transformation to apply first

        Returns:
            Transform: the composition of the two transformations
        """
        transform = self.transform @ other.transform
        source_origin = other.source_origin
        destination_origin = self.apply(other.destination_origin) if other.destination_origin is not None else None
        return Transform(transform, source_origin, destination_origin)

    def _standard_input(self) -> Quantity:
        """Construct standard input points (1,0), (0,1) and (0,0) with the source unit of this transform."""
        return Quantity(((1.0, 0.0, 0.0), (0.0, 1.0, 0.0)), self.source_origin)

    @classmethod
    def identity(cls):
        return Transform()


def place(out_shape: tuple[int, ...], out_pixel_size: Quantity, source: np.ndarray, offset: Optional[Quantity] = None,
          out: Optional[np.ndarray] = None):
    """Takes a source array and places it in an otherwise empty array of specified shape and pixel size.

    The source array must have a pixel_size property (see set_pixel_size).
    It is scaled (using area interpolation) so that the pixel size matches that of the output.
    Then, the source array is placed into the output array at the position given by offset,
    where `offset=None` or `offset=(0.0, 0.0) * pixel_size` corresponds to centering the source image
    with respect to the output array.
    Parts of the source array that extend beyond the output are cropped, and parts of the array that are not
    covered by the source array are zero padded.

    Note: this function currently works for 2-D inputs only

    Args:

    """
    out_extent = out_pixel_size * np.array(out_shape)
    transform = Transform(destination_origin=offset)
    return project(source, out_extent=out_extent, out_shape=out_shape, transform=transform, out=out)


def project(
        source: np.ndarray, *,
        source_extent: Optional[ExtentType] = None,
        transform: Optional[Transform] = None,
        out: Optional[np.ndarray] = None,
        out_extent: Optional[ExtentType] = None,
        out_shape: Optional[tuple[int, ...]] = None) -> np.ndarray:
    """Projects the input image onto an array with specified shape and resolution.

    The input image is scaled so that the pixel sizes match those of the output,
    and cropped/zero-padded so that the data shape matches that of the output.
    Optionally, an additional transformation can be specified, e. g., to scale or translate the source image.
    This transformation is specified as a 2x3 transformation matrix in homogeneous coordinates.

    Args:
        source (np.ndarray): input image.
            Must have the pixel_size set (see set_pixel_size)
        transform: transformation to appy to the source image before placing it in the output
        out (np.ndarray): optional array where the output image is stored in.
            If specified, `out_shape` is ignored.

    Returns:
        np.ndarray: the projected image (`out` if specified, otherwise a new array)
    """
    transform = transform if transform is not None else Transform()
    if out is not None:
        if out_shape is not None and out_shape != out.shape:
            raise ValueError("out_shape and out.shape must match. Note that out_shape may be omitted")
        if out.dtype != source.dtype:
            raise ValueError("out and source must have the same dtype")
        out_shape = out.shape
        out_extent = out_extent or get_extent(out)
    if out_shape is None:
        raise ValueError("Either out_shape or out must be specified")
    if out_extent is None:
        raise ValueError("Either out_extent or the pixel_size metadata of out must be specified")
    source_extent = source_extent if source_extent is not None else get_extent(source)
    source_ps = source_extent / np.array(source.shape)
    out_ps = out_extent / np.array(out_shape)

    t = transform.cv2_matrix(source.shape, source_ps, out_shape, out_ps)
    # swap x and y in matrix and size, since cv2 uses the (x,y) convention.
    out_size = (out_shape[1], out_shape[0])
    if (source.dtype == np.complex128) or (source.dtype == np.complex64):
        if out is None:
            out = np.zeros(out_shape, dtype=source.dtype)
        # real part
        out.real = cv2.warpAffine(source.real, t, out_size, flags=cv2.INTER_NEAREST,
                                  borderMode=cv2.BORDER_CONSTANT, borderValue=(0.0,))
        # imaginary part
        out.imag = cv2.warpAffine(source.imag, t, out_size, flags=cv2.INTER_NEAREST,
                                  borderMode=cv2.BORDER_CONSTANT, borderValue=(0.0,))

    else:
        dst = cv2.warpAffine(source, t, out_size, dst=out, flags=cv2.INTER_NEAREST,
                             borderMode=cv2.BORDER_CONSTANT, borderValue=(0.0,))
        if out is not None and out is not dst:
            raise ValueError("OpenCV did not use the specified output array. This should not happen.")
        out = dst
    return set_pixel_size(out, out_ps)


def set_pixel_size(data: ArrayLike, pixel_size: Optional[Quantity]) -> np.ndarray:
    """
    Sets the pixel size metadata for the given data array.

    Args:
        data (ArrayLike): The input data array.
        pixel_size (Optional[Quantity]): The pixel size to be set. When a single-element pixel size is given,
            it is broadcasted to all dimensions of the data array.
            Passing None sets the pixel size metadata to None.

    Returns:
        np.ndarray: The modified data array with the pixel size metadata.

    Usage:
    >>> data = np.array([[1, 2], [3, 4]])
    >>> pixel_size = 0.1 * u.m
    >>> modified_data = set_pixel_size(data, pixel_size)
    """
    data = np.array(data)

    if pixel_size is not None and pixel_size.size == 1:
        pixel_size = pixel_size * np.ones(data.ndim)

    data.dtype = np.dtype(data.dtype, metadata={'pixel_size': pixel_size})
    return data


def get_pixel_size(data: np.ndarray) -> Optional[Quantity]:
    """
    Extracts the pixel size metadata from the data array.

    Args:
        data (np.ndarray): The input data array or Quantity.

    Returns:
        OptionalQuantity]: The pixel size metadata, or None if no pixel size metadata is present.

    Usage:
    >>> import astropy.units as u
    >>> import numpy as np
    >>> data = set_pixel_size(((1, 2), (3, 4)), 5 * u.um)
    >>> pixel_size = get_pixel_size(data)

    TODO: return 1 if no pixel size is present
    """
    metadata = data.dtype.metadata
    if metadata is None:
        return None
    return data.dtype.metadata.get('pixel_size', None)


def get_extent(data: np.ndarray) -> Quantity:
    """
    Extracts the extent from the data array.
    The extent is always equal to `shape * pixel_size`.
    """
    pixel_size = get_pixel_size(data)
    if pixel_size is None:
        return Quantity(data.shape)
    return data.shape * pixel_size


def set_extent(data: np.ndarray, extent: ExtentType) -> np.ndarray:
    """
    Sets the extent metadata for the given data array.

    Args:
        data (ArrayLike): The input data array.
        extent (ArrayLike): The extent to be set. When a single-element extent is given,
            it is broadcasted to all dimensions of the data array.
            Passing None sets the extent metadata to None.

    Returns:
        np.ndarray: The modified data array with the extent metadata.

    Usage:
    >>> data = np.array([[1, 2], [3, 4]])
    >>> extent = 0.1 * u.m
    >>> modified_data = set_extent(data, extent)
    """
    return set_pixel_size(data, extent / np.array(data.shape))



import warnings

import numpy as np

from .context import Context

try:
    import OpenGL.GL as GL
    from OpenGL.GL import glGenTextures, glBindTexture, glTexImage2D, glTexSubImage2D, glTexImage1D, glTexSubImage1D, \
        glTexParameteri, glActiveTexture, glDeleteTextures, glGetTextureImage, glPixelStorei
except AttributeError:
    warnings.warn("OpenGL not found, SLM will not work"),


class Texture:
    def __init__(self, slm, texture_type=GL.GL_TEXTURE_2D):
        self.context = Context(slm)
        self.handle = glGenTextures(1)
        self.type = texture_type
        self.synchronized = False  # self.data is not yet synchronized with texture in GPU memory
        self._data_shape = None  # current size of the texture, to see if we need to make a new texture or
        # overwrite the exiting one

        # create a single pixel texture as default (also activates the OpenGL context and binds the texture
        self.set_data(0)

        # set wrapping and interpolation options
        glTexParameteri(self.type, GL.GL_TEXTURE_WRAP_S, GL.GL_REPEAT)
        glTexParameteri(self.type, GL.GL_TEXTURE_WRAP_T, GL.GL_REPEAT)
        glTexParameteri(self.type, GL.GL_TEXTURE_WRAP_R, GL.GL_REPEAT)
        glTexParameteri(self.type, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)
        glTexParameteri(self.type, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)

    def __del__(self):
        with self.context as slm:
            if slm:
                glDeleteTextures(1, [self.handle])

    def _bind(self, idx):
        """ Bind texture to texture unit idx. Assumes that the OpenGL context is already active."""
        glActiveTexture(GL.GL_TEXTURE0 + idx)
        glBindTexture(self.type, self.handle)

    def set_data(self, value):
        """ Set texture data.

        The texture data is directly copied to the GPU memory,
         so the original data array can be modified or deleted.
        """
        value = np.array(value, dtype=np.float32, order='C', copy=False)

        with self.context:
            glBindTexture(self.type, self.handle)
            glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 4)  # alignment is at least four bytes since we use float32
            (internal_format, data_format, data_type) = (GL.GL_R32F, GL.GL_RED, GL.GL_FLOAT)

            if self.type == GL.GL_TEXTURE_1D:
                # check if data has the correct dimension, convert scalars to arrays of correct dimension
                if value.ndim == 0:
                    value = value.reshape((1,))
                elif value.ndim != 1:
                    raise ValueError("Data should be a 1-d array or a scalar")
                if value.shape != self._data_shape:
                    # create a new texture
                    glTexImage1D(GL.GL_TEXTURE_1D, 0, internal_format, value.shape[0], 0, data_format, data_type, value)
                    self._data_shape = value.shape
                else:
                    # overwrite existing texture
                    glTexSubImage1D(GL.GL_TEXTURE_1D, 0, 0, value.shape[0], data_format, data_type, value)

            elif self.type == GL.GL_TEXTURE_2D:
                if value.ndim == 0:
                    value = value.reshape((1, 1))
                elif value.ndim != 2:
                    raise ValueError("Data should be a 2-D array or a scalar")
                if value.shape != self._data_shape:
                    glTexImage2D(GL.GL_TEXTURE_2D, 0, internal_format, value.shape[1], value.shape[0], 0,
                                 data_format, data_type, value)
                    self._data_shape = value.shape
                else:
                    glTexSubImage2D(GL.GL_TEXTURE_2D, 0, 0, 0, value.shape[1], value.shape[0], data_format,
                                    data_type, value)
            else:
                raise ValueError("Texture type not supported")

    def get_data(self):
        with self.context:
            data = np.empty(self._data_shape, dtype='float32')
            glGetTextureImage(self.handle, 0, GL.GL_RED, GL.GL_FLOAT, data.size * 4, data)
            return data
